from typing import Dict, Optional
from sqlalchemy import Column, Numeric, Integer, ForeignKey, select
from sqlalchemy.orm import relationship, backref, object_session
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.hybrid import hybrid_method

from glycresoft.serialize.analysis import BoundToAnalysis

from glycresoft.serialize.chromatogram import ChromatogramSolution, ChromatogramWrapper, CompoundMassShift, MissingChromatogramError

from glycresoft.serialize.tandem import (
    GlycopeptideSpectrumCluster,
    GlycopeptideSpectrumMatch,
    GlycopeptideSpectrumSolutionSet,
)

from glycresoft.serialize.hypothesis import Glycopeptide

from glycresoft.tandem.chromatogram_mapping import TandemAnnotatedChromatogram
from glycresoft.chromatogram_tree import GlycopeptideChromatogram
from glycresoft.chromatogram_tree.utils import ArithmeticMapping

from glycresoft.serialize.base import Base


class IdentifiedStructure(BoundToAnalysis, ChromatogramWrapper):
    @declared_attr
    def id(self):
        return Column(Integer, primary_key=True)

    @declared_attr
    def chromatogram_solution_id(self):
        return Column(Integer, ForeignKey(ChromatogramSolution.id, ondelete="CASCADE"), index=True)

    def has_chromatogram(self):
        return self.chromatogram_solution_id is not None

    @declared_attr
    def chromatogram(self):
        return relationship(ChromatogramSolution, lazy="select")

    def get_chromatogram(self):
        return self.chromatogram.get_chromatogram()

    @property
    def total_signal(self):
        try:
            return self.chromatogram.total_signal
        except AttributeError:
            return 0

    @property
    def composition(self):
        raise NotImplementedError()

    @property
    def entity(self):
        raise NotImplementedError()

    def bisect_mass_shift(self, mass_shift):
        chromatogram = self._get_chromatogram()
        new_mass_shift, new_no_mass_shift = chromatogram.bisect_mass_shift(mass_shift)

        for chrom in (new_mass_shift, new_no_mass_shift):
            chrom.entity = self.entity
            chrom.composition = self.entity
            chrom.glycan_composition = self.glycan_composition
        return new_mass_shift, new_no_mass_shift


class AmbiguousGlycopeptideGroup(Base, BoundToAnalysis):
    __tablename__ = "AmbiguousGlycopeptideGroup"

    id = Column(Integer, primary_key=True)

    def __repr__(self):
        return "AmbiguousGlycopeptideGroup(%d)" % (self.id,)

    @classmethod
    def serialize(cls, members, session, analysis_id, *args, **kwargs):
        inst = cls(analysis_id=analysis_id)
        session.add(inst)
        session.flush()
        for member in members:
            member.ambiguous_id = inst.id
            session.add(member)
        session.flush()
        return inst

    def __iter__(self):
        return iter(self.members)


class IdentifiedGlycopeptide(Base, IdentifiedStructure):
    __tablename__ = "IdentifiedGlycopeptide"

    structure_id = Column(Integer, ForeignKey(Glycopeptide.id, ondelete="CASCADE"), index=True)

    structure = relationship(Glycopeptide)

    ms1_score = Column(Numeric(8, 7, asdecimal=False), index=True)
    ms2_score = Column(Numeric(12, 6, asdecimal=False), index=True)
    q_value = Column(Numeric(8, 7, asdecimal=False), index=True)

    spectrum_cluster_id = Column(Integer, ForeignKey(GlycopeptideSpectrumCluster.id, ondelete="CASCADE"), index=True)

    spectrum_cluster = relationship(GlycopeptideSpectrumCluster, backref=backref("owners", lazy="dynamic"))

    ambiguous_id = Column(Integer, ForeignKey(AmbiguousGlycopeptideGroup.id, ondelete="CASCADE"), index=True)

    shared_with = relationship(AmbiguousGlycopeptideGroup, backref=backref("members", lazy="dynamic"))

    @property
    def score_set(self):
        """The :class:`~.ScoreSet` of the best MS/MS match

        Returns
        -------
        :class:`~.ScoreSet`
        """
        if not self.is_multiscore():
            return None
        best_match = self._best_spectrum_match
        if best_match is None:
            return None
        return best_match.score_set

    @property
    def localizations(self):
        return self.best_spectrum_match.localizations

    @property
    def best_spectrum_match(self) -> GlycopeptideSpectrumMatch:
        return self._best_spectrum_match

    @property
    def q_value_set(self):
        """The :class:`~.FDRSet` of the best MS/MS match

        Returns
        -------
        :class:`~.FDRSet`
        """
        if not self.is_multiscore():
            return None
        best_match = self._best_spectrum_match
        if best_match is None:
            return None
        return best_match.q_value_set

    def is_multiscore(self) -> bool:
        """Check whether this match collection has been produced by summarizing a multi-score
        match, rather than a single score match.

        Returns
        -------
        bool
        """
        return self.spectrum_cluster.is_multiscore()

    @property
    def _best_spectrum_match(self):
        summary: IdentifiedGlycopeptideSummary = self._summary
        if summary is not None and summary.best_spectrum_match_id is not None:
            session = object_session(self)
            return session.query(GlycopeptideSpectrumMatch).get(summary.best_spectrum_match_id)

        is_multiscore = self.is_multiscore()
        best_match = None
        if is_multiscore:
            best_score = 0.0
            best_q_value = 1.0
        else:
            best_score = 0.0
        for solution_set in self.spectrum_cluster:
            try:
                match = solution_set.solution_for(self.structure)
                if is_multiscore:
                    q_value = match.q_value
                    if q_value <= best_q_value:
                        q_delta = abs(best_q_value - q_value)
                        best_q_value = q_value
                        if q_delta > 0.001:
                            best_score = match.score
                            best_match = match
                        else:
                            score = match.score
                            if score > best_score:
                                best_score = score
                                best_match = match
                else:
                    score = match.score
                    if score > best_score:
                        best_score = score
                        best_match = match
            except KeyError:
                continue
        return best_match

    @hybrid_method
    def is_multiply_glycosylated(self):
        return self.structure.is_multiply_glycosylated()

    @is_multiply_glycosylated.expression
    def is_multiply_glycosylated(self):
        expr = (
            select([Glycopeptide.is_multiply_glycosylated()])
            .where(Glycopeptide.id == IdentifiedGlycopeptide.structure_id)
            .label("is_multiply_glycosylated")
        )
        return expr

    @property
    def glycan_composition(self):
        return self.structure.glycan_composition

    @property
    def protein_relation(self):
        return self.structure.protein_relation

    @property
    def start_position(self):
        return self.protein_relation.start_position

    @property
    def end_position(self):
        return self.protein_relation.end_position

    def overlaps(self, other):
        return self.protein_relation.overlaps(other.protein_relation)

    def spans(self, position):
        return position in self.protein_relation

    @property
    def spectrum_matches(self):
        return self.spectrum_cluster.spectrum_solutions

    @classmethod
    def serialize(
        cls,
        obj,
        session,
        chromatogram_solution_id,
        tandem_cluster_id,
        analysis_id,
        build_summary: bool = False,
        *args,
        **kwargs,
    ):
        inst = cls(
            chromatogram_solution_id=chromatogram_solution_id,
            spectrum_cluster_id=tandem_cluster_id,
            analysis_id=analysis_id,
            q_value=obj.q_value,
            ms2_score=obj.ms2_score,
            ms1_score=obj.ms1_score,
            structure_id=obj.structure.id,
        )
        if build_summary:
            inst._build_summary()
        session.add(inst)
        session.flush()
        return inst

    @property
    def tandem_solutions(self):
        return self.spectrum_cluster.spectrum_solutions

    def mass_shift_tandem_solutions(self):
        mapping = ArithmeticMapping()
        for sm in self.tandem_solutions:
            try:
                mapping[sm.solution_for(self.structure).mass_shift] += 1
            except KeyError:
                continue
        return mapping

    def get_chromatogram(self, *args, **kwargs):
        chromatogram = self.chromatogram.convert(*args, **kwargs)
        chromatogram.chromatogram = chromatogram.chromatogram.clone(GlycopeptideChromatogram)
        structure = self.structure.convert()
        chromatogram.chromatogram.entity = structure
        return chromatogram

    @classmethod
    def bulk_convert(
        cls,
        iterable,
        expand_shared_with: bool = True,
        mass_shift_cache: Optional[Dict] = None,
        scan_cache: Optional[Dict] = None,
        structure_cache: Optional[Dict] = None,
        peptide_relation_cache: Optional[Dict] = None,
        shared_identification_cache: Optional[Dict] = None,
        min_q_value: float = 0.2,
        *args,
        **kwargs,
    ):
        session = object_session(iterable[0])
        if mass_shift_cache is None:
            mass_shift_cache = {m.id: m.convert() for m in session.query(CompoundMassShift).all()}
        if scan_cache is None:
            scan_cache = {}
        if structure_cache is None:
            structure_cache = {}
        if peptide_relation_cache is None:
            peptide_relation_cache = {}
        if shared_identification_cache is None:
            shared_identification_cache = {}
        result = [
            obj.convert(
                expand_shared_with=expand_shared_with,
                mass_shift_cache=mass_shift_cache,
                scan_cache=scan_cache,
                structure_cache=structure_cache,
                peptide_relation_cache=peptide_relation_cache,
                min_q_value=min_q_value,
                *args,
                **kwargs,
            )
            for obj in iterable
        ]
        return result

    def convert(
        self,
        expand_shared_with: bool = True,
        mass_shift_cache: Optional[Dict] = None,
        scan_cache: Optional[Dict] = None,
        structure_cache: Optional[Dict] = None,
        peptide_relation_cache: Optional[Dict] = None,
        shared_identification_cache: Optional[Dict] = None,
        min_q_value: float = 0.2,
        *args,
        **kwargs,
    ):
        # bind this name late to avoid circular import error
        from glycresoft.tandem.glycopeptide.identified_structure import (
            IdentifiedGlycopeptide as MemoryIdentifiedGlycopeptide,
        )

        session = object_session(self)
        if mass_shift_cache is None:
            mass_shift_cache = {m.id: m.convert() for m in session.query(CompoundMassShift).all()}
        if shared_identification_cache is None:
            shared_identification_cache = dict()
        if scan_cache is None:
            scan_cache = dict()
        if structure_cache is None:
            structure_cache = dict()
        if peptide_relation_cache is None:
            peptide_relation_cache = dict()
        if expand_shared_with and self.shared_with.members.count() > 1:
            stored = list(self.shared_with)
            converted = []
            for idgp in stored:
                if idgp.id in shared_identification_cache:
                    converted.append(shared_identification_cache[idgp.id])
                else:
                    tmp = idgp.convert(
                        expand_shared_with=False,
                        mass_shift_cache=mass_shift_cache,
                        scan_cache=scan_cache,
                        structure_cache=structure_cache,
                        peptide_relation_cache=peptide_relation_cache,
                        shared_identification_cache=shared_identification_cache,
                        min_q_value=min_q_value,
                        *args,
                        **kwargs,
                    )
                    shared_identification_cache[idgp.id] = tmp
                    converted.append(tmp)
            for i in range(len(stored)):
                converted[i].shared_with = converted[:i] + converted[i + 1 :]
            for i, member in enumerate(stored):
                if member.id == self.id:
                    return converted[i]
        else:
            spectrum_matches = GlycopeptideSpectrumSolutionSet.bulk_load_from_clusters(
                session,
                [self.spectrum_cluster_id],
                mass_shift_cache=mass_shift_cache,
                scan_cache=scan_cache,
                structure_cache=structure_cache,
                peptide_relation_cache=peptide_relation_cache,
                min_q_value=min_q_value,
            )[self.spectrum_cluster_id]
            if structure_cache is not None:
                if self.structure_id in structure_cache:
                    structure, _ = structure_cache[self.structure_id]
                else:
                    structure = Glycopeptide.bulk_load(
                        session,
                        [self.structure_id],
                        peptide_relation_cache=peptide_relation_cache,
                        structure_cache=structure_cache,
                    )[0]
            else:
                structure = self.structure.convert()

            chromatogram = self.chromatogram
            if chromatogram is not None:
                chromatogram = self.chromatogram.convert(*args, **kwargs)
                chromatogram.chromatogram = TandemAnnotatedChromatogram(
                    chromatogram.chromatogram.clone(GlycopeptideChromatogram)
                )
                chromatogram.chromatogram.tandem_solutions.extend(spectrum_matches)
                chromatogram.chromatogram.entity = structure

            inst = MemoryIdentifiedGlycopeptide(structure, spectrum_matches, chromatogram)
            inst.id = self.id
            shared_identification_cache[self.id] = inst
            return inst

    @property
    def composition(self):
        return self.structure.convert()

    @property
    def entity(self):
        return self.structure.convert()

    def __repr__(self):
        return "DB" + repr(self.convert())

    def __str__(self):
        return str(self.structure)

    def has_scan(self, scan_id: str) -> bool:
        return any([sset.scan.scan_id == scan_id for sset in self.tandem_solutions])

    def get_scan(self, scan_id: str):
        for sset in self.tandem_solutions:
            if sset.scan.scan_id == scan_id:
                return sset
        raise KeyError(scan_id)

    @declared_attr
    def _summary(self):
        return relationship("IdentifiedGlycopeptideSummary", lazy="joined", uselist=False)

    @property
    def total_signal(self):
        summary = self._summary
        if summary is not None:
            return summary.total_signal
        try:
            return self.chromatogram.total_signal
        except AttributeError:
            return 0

    @property
    def apex_time(self):
        summary = self._summary
        if summary is not None and summary.apex_time is not None:
            return summary.apex_time
        return super().apex_time

    @property
    def start_time(self):
        summary = self._summary
        if summary is not None and summary.start_time is not None:
            return summary.start_time
        return super().start_time

    @property
    def end_time(self):
        summary = self._summary
        if summary is not None and summary.end_time is not None:
            return summary.end_time
        return super().end_time

    @property
    def weighted_neutral_mass(self):
        summary = self._summary
        if summary is not None and summary.weighted_neutral_mass is not None:
            return summary.weighted_neutral_mass
        return super().weighted_neutral_mass

    def _build_summary(self):
        if self._summary is not None:
            return
        session = object_session(self)

        try:
            apex_time = self.apex_time
            total_signal = self.total_signal
            start_time = self.start_time
            end_time = self.end_time
        except MissingChromatogramError:
            apex_time = total_signal = start_time = end_time = None
        best_spectrum_match_id = self.best_spectrum_match.id

        summary = IdentifiedGlycopeptideSummary(
            id=self.id,
            weighted_neutral_mass=self.weighted_neutral_mass,
            apex_time=apex_time,
            total_signal=total_signal,
            start_time=start_time,
            end_time=end_time,
            best_spectrum_match_id=best_spectrum_match_id,
        )
        session.add(summary)
        return summary


class IdentifiedGlycopeptideSummary(Base):
    __tablename__ = "IdentifiedGlycopeptideSummary"

    id = Column(Integer, ForeignKey(IdentifiedGlycopeptide.id), primary_key=True)

    weighted_neutral_mass = Column(Numeric(12, 6, asdecimal=False), index=True)
    apex_time = Column(Numeric(12, 6, asdecimal=False), index=True)
    total_signal = Column(Numeric(12, 6, asdecimal=False), index=True)
    start_time = Column(Numeric(12, 6, asdecimal=False), index=True)
    end_time = Column(Numeric(12, 6, asdecimal=False), index=True)
    best_spectrum_match_id = Column(
        Integer, ForeignKey(GlycopeptideSpectrumMatch.id, ondelete="CASCADE"), index=True, nullable=True
    )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(id={self.id}, weighted_neutral_mass={self.weighted_neutral_mass}, apex_time={self.apex_time}, "
            f"total_signal={self.total_signal}, start_time={self.start_time}, end_time={self.end_time}, "
            f"best_spectrum_match_id={self.best_spectrum_match_id})"
        )



import bisect
import itertools

from multiprocessing import Process, Queue, Event, cpu_count

from typing import List, Optional, Set, Tuple

from lxml.etree import XMLSyntaxError

from ms_deisotope.peak_dependency_network.intervals import Interval, IntervalTreeNode

from glycopeptidepy import enzyme
from .utils import slurp
from .uniprot import (uniprot, get_uniprot_accession, UniprotProteinDownloader, UniprotProteinXML, Empty)

from glypy.composition import formula
from glycopeptidepy.io import annotation
from glycopeptidepy.structure import sequence, modification, residue
from glycopeptidepy.structure.modification import ModificationRule
from glycopeptidepy.algorithm import PeptidoformGenerator
from glycopeptidepy import PeptideSequence

from glycresoft.task import TaskBase
from glycresoft.serialize import DatabaseBoundOperation
from glycresoft.serialize.hypothesis.peptide import Peptide, Protein

from six import string_types as basestring


RestrictedModificationTable = modification.RestrictedModificationTable
combinations = itertools.combinations
product = itertools.product
chain_iterable = itertools.chain.from_iterable

SequenceLocation = modification.SequenceLocation


def span_test(site_list, start, end):
    i = bisect.bisect_left(site_list, start)
    after_start = site_list[i:]
    out = []
    for j in after_start:
        if j < end:
            out.append(j)
        else:
            break
    return out


def n_glycan_sequon_sites(peptide: Peptide, protein: Protein, use_local_sequence: bool=False,
                          include_cysteine: bool = False, allow_modified=frozenset()):
    sites = set()
    sites |= set(site - peptide.start_position for site in span_test(
        protein.n_glycan_sequon_sites, peptide.start_position, peptide.end_position))
    if use_local_sequence:
        sites |= set(sequence.find_n_glycosylation_sequons(
            peptide.modified_peptide_sequence, allow_modified=allow_modified, include_cysteine=include_cysteine))
    return sorted(sites)


def o_glycan_sequon_sites(peptide: Peptide, protein: Protein=None, use_local_sequence: bool=False):
    sites = set()
    sites |= set(site - peptide.start_position for site in span_test(
        protein.o_glycan_sequon_sites, peptide.start_position, peptide.end_position))
    if use_local_sequence:
        sites |= set(sequence.find_o_glycosylation_sequons(
            peptide.modified_peptide_sequence))
    return sorted(sites)


def gag_sequon_sites(peptide: Peptide, protein: Protein = None, use_local_sequence: bool = False):
    sites = set()
    sites |= set(site - peptide.start_position for site in span_test(
        protein.glycosaminoglycan_sequon_sites, peptide.start_position, peptide.end_position))
    if use_local_sequence:
        sites = sequence.find_glycosaminoglycan_sequons(
            peptide.modified_peptide_sequence)
    return sorted(sites)


def get_base_peptide(peptide_obj):
    if isinstance(peptide_obj, Peptide):
        return PeptideSequence(peptide_obj.base_peptide_sequence)
    else:
        return PeptideSequence(str(peptide_obj))


class PeptidePermuter(PeptidoformGenerator):
    constant_modifications: List[ModificationRule]
    variable_modifications: List[ModificationRule]
    max_variable_modifications: int

    n_term_modifications: List[ModificationRule]
    c_term_modifications: List[ModificationRule]

    def prepare_peptide(self, sequence):
        return get_base_peptide(sequence)

    @classmethod
    def peptide_permutations(cls,
                             sequence,
                             constant_modifications: List[ModificationRule],
                             variable_modifications: List[ModificationRule],
                             max_variable_modifications: int=4):
        inst = cls(constant_modifications, variable_modifications,
                   max_variable_modifications)
        return inst(sequence)


peptide_permutations = PeptidePermuter.peptide_permutations


def cleave_sequence(sequence,
                    protease: enzyme.Protease,
                    missed_cleavages: int=2,
                    min_length: int=6,
                    max_length: int=60,
                    semispecific: bool=False):
    peptide: str
    start: int
    end: int
    missed: int
    for peptide, start, end, missed in protease.cleave(sequence, missed_cleavages=missed_cleavages,
                                                       min_length=min_length, max_length=max_length,
                                                       semispecific=semispecific):
        if "X" in peptide:
            continue
        yield peptide, start, end, missed


class ProteinDigestor(TaskBase):

    protease: enzyme.Protease
    constant_modifications: List[ModificationRule]
    variable_modifications: List[ModificationRule]
    peptide_permuter: PeptidePermuter
    min_length: int
    max_length: int
    max_missed_cleavages: int
    variable_signal_peptide: int
    max_variable_modifications: int
    semispecific: bool
    include_cysteine_n_glycosylation: bool
    require_glycosylation_sites: bool

    def __init__(self, protease, constant_modifications=None, variable_modifications=None,
                 max_missed_cleavages: int=2, min_length: int=6, max_length: int=60, semispecific: bool=False,
                 max_variable_modifications=None, require_glycosylation_sites: bool=False,
                 include_cysteine_n_glycosylation: bool=False):
        if constant_modifications is None:
            constant_modifications = []
        if variable_modifications is None:
            variable_modifications = []
        self.protease = self._prepare_protease(protease)
        self.constant_modifications = constant_modifications
        self.variable_modifications = variable_modifications
        self.peptide_permuter = PeptidePermuter(
            self.constant_modifications,
            self.variable_modifications,
            max_variable_modifications=max_variable_modifications)
        self.max_missed_cleavages = max_missed_cleavages
        self.min_length = min_length
        self.max_length = max_length
        self.semispecific = semispecific
        self.include_cysteine_n_glycosylation = include_cysteine_n_glycosylation
        self.max_variable_modifications = max_variable_modifications
        self.require_glycosylation_sites = require_glycosylation_sites

    def _prepare_protease(self, protease):
        if isinstance(protease, enzyme.Protease):
            pass
        elif isinstance(protease, basestring):
            protease = enzyme.Protease(protease)
        elif isinstance(protease, (list, tuple)):
            protease = enzyme.Protease.combine(*protease)
        return protease

    def cleave(self, sequence: Protein):
        return cleave_sequence(sequence, self.protease, self.max_missed_cleavages,
                               min_length=self.min_length, max_length=self.max_length,
                               semispecific=self.semispecific)

    def digest(self, protein: Protein):
        sequence = protein.protein_sequence
        try:
            size = len(protein)
        except residue.UnknownAminoAcidException:
            size = len(sequence)
        for peptide, start, end, n_missed_cleavages in self.cleave(sequence):
            if end - start > self.max_length:
                continue
            protein_n_term = start == 0
            protein_c_term = end == size
            for inst in self.modify_string(
                    peptide, protein_n_term=protein_n_term, protein_c_term=protein_c_term):
                inst.count_missed_cleavages = n_missed_cleavages
                inst.start_position = start
                inst.end_position = end
                yield inst

    def modify_string(self, peptide: Peptide, protein_n_term: bool=False, protein_c_term: bool=False):
        base_peptide = str(peptide)
        for modified_peptide, n_variable_modifications in self.peptide_permuter(
                peptide, protein_n_term=protein_n_term, protein_c_term=protein_c_term):
            formula_string = formula(modified_peptide.total_composition())
            # formula_string = ""
            inst = Peptide(
                base_peptide_sequence=base_peptide,
                modified_peptide_sequence=str(modified_peptide),
                count_missed_cleavages=-1,
                count_variable_modifications=n_variable_modifications,
                sequence_length=len(modified_peptide),
                start_position=-1,
                end_position=-1,
                calculated_mass=modified_peptide.mass,
                formula=formula_string)
            yield inst

    def process_protein(self, protein_obj: Protein):
        protein_id = protein_obj.id
        hypothesis_id = protein_obj.hypothesis_id

        for peptide in self.digest(protein_obj):
            peptide.protein_id = protein_id
            peptide.hypothesis_id = hypothesis_id
            peptide.peptide_score = 0
            peptide.peptide_score_type = 'null_score'
            n_glycosites = n_glycan_sequon_sites(
                peptide, protein_obj, include_cysteine=self.include_cysteine_n_glycosylation)
            o_glycosites = o_glycan_sequon_sites(peptide, protein_obj)
            gag_glycosites = gag_sequon_sites(peptide, protein_obj)
            if self.require_glycosylation_sites:
                if (len(n_glycosites) + len(o_glycosites) + len(gag_glycosites)) == 0:
                    continue
            peptide.count_glycosylation_sites = len(n_glycosites)
            peptide.n_glycosylation_sites = sorted(n_glycosites)
            peptide.o_glycosylation_sites = sorted(o_glycosites)
            peptide.gagylation_sites = sorted(gag_glycosites)
            yield peptide

    def __call__(self, protein_obj):
        return self.process_protein(protein_obj)

    @classmethod
    def digest_protein(cls, sequence, protease: enzyme.Protease, constant_modifications: List[ModificationRule] = None,
                       variable_modifications: List[ModificationRule] = None, max_missed_cleavages: int=2,
                       min_length: int=6, max_length: int=60, semispecific: bool=False):
        inst = cls(
            protease, constant_modifications, variable_modifications,
            max_missed_cleavages, min_length, max_length, semispecific)
        if isinstance(sequence, basestring):
            sequence = Protein(protein_sequence=sequence)
        return inst(sequence)


digest = ProteinDigestor.digest_protein


class ProteinDigestingProcess(Process):
    process_name = "protein-digest-worker"

    def __init__(self, connection, hypothesis_id, input_queue, digestor, done_event=None,
                 chunk_size=5000, message_handler=None):
        Process.__init__(self)
        self.connection = connection
        self.input_queue = input_queue
        self.hypothesis_id = hypothesis_id
        self.done_event = done_event
        self.digestor = digestor
        self.chunk_size = chunk_size
        self.message_handler = message_handler

    def task(self):
        database = DatabaseBoundOperation(self.connection)
        session = database.session
        has_work = True

        digestor = self.digestor
        acc = []
        if self.message_handler is None:
            self.message_handler = lambda x: None
        while has_work:
            try:
                work_items = self.input_queue.get(timeout=5)
                if work_items is None:
                    has_work = False
                    continue
            except Exception:
                if self.done_event.is_set():
                    has_work = False
                continue
            proteins = slurp(session, Protein, work_items, flatten=False)
            acc = []

            threshold_size = 3000

            for protein in proteins:
                size = len(protein.protein_sequence)
                if size > threshold_size:
                    self.message_handler("...... Started digesting %s (%d)" % (protein.name, size))
                i = 0
                for peptide in digestor.process_protein(protein):
                    acc.append(peptide)
                    i += 1
                    if len(acc) > self.chunk_size:
                        session.bulk_save_objects(acc)
                        session.commit()
                        acc = []
                    if i % 10000 == 0:
                        self.message_handler(
                            "...... Digested %d peptides from %r (%d)" % (
                                i, protein.name, size))
                if size > threshold_size:
                    self.message_handler("...... Finished digesting %s (%d)" % (protein.name, size))
            session.bulk_save_objects(acc)
            session.commit()
            acc = []
        if acc:
            session.bulk_save_objects(acc)
            session.commit()
            acc = []

    def run(self):
        new_name = getattr(self, 'process_name', None)
        if new_name is not None:
            TaskBase().try_set_process_name(new_name)
        self.task()


class MultipleProcessProteinDigestor(TaskBase):
    def __init__(self, connection, hypothesis_id, protein_ids, digestor, n_processes=4):
        self.connection = connection
        self.hypothesis_id = hypothesis_id
        self.protein_ids = protein_ids
        self.digestor = digestor
        self.n_processes = n_processes

    def run(self):
        logger = self.ipc_logger()
        input_queue = Queue(2 * self.n_processes)
        done_event = Event()
        processes = [
            ProteinDigestingProcess(
                self.connection, self.hypothesis_id, input_queue,
                self.digestor, done_event=done_event,
                message_handler=logger.sender()) for i in range(
                self.n_processes)
        ]
        protein_ids = self.protein_ids
        i = 0
        n = len(protein_ids)
        if n <= self.n_processes:
            chunk_size = 1
        elif n < 100:
            chunk_size = 2
        else:
            chunk_size = int(n / 100.0)
        interval = 30
        for process in processes:
            input_queue.put(protein_ids[i:(i + chunk_size)])
            i += chunk_size
            process.start()

        last = i
        while i < n:
            input_queue.put(protein_ids[i:(i + chunk_size)])
            i += chunk_size
            if i - last > interval:
                self.log("... Dealt Proteins %d-%d %0.2f%%" % (
                    i - chunk_size, min(i, n), (min(i, n) / float(n)) * 100))
                last = i

                error_occurred = False
                for process in processes:
                    if process.exitcode is not None and process.exitcode != 0:
                        error_occurred = True

                if error_occurred:
                    self.error("An error occurred while digesting proteins.")
                    done_event.set()
                    for process in processes:
                        if process.is_alive():
                            process.terminate()
                    logger.stop()
                    raise ValueError(
                        "One or more worker processes exited with an error.")

        done_event.set()
        for process in processes:
            process.join()
            if process.exitcode != 0:
                raise ValueError("One or more worker processes exited with an error.")
        logger.stop()


class PeptideInterval(Interval):
    def __init__(self, peptide):
        Interval.__init__(self, peptide.start_position, peptide.end_position, [peptide])
        self.data['peptide'] = str(peptide)


class ProteinSplitter(TaskBase):

    constant_modifications: List[ModificationRule]
    variable_modifications: List[ModificationRule]
    min_length: int
    variable_signal_peptide: int
    include_cysteine_n_glycosylation: bool

    def __init__(self, constant_modifications=None, variable_modifications=None,
                 min_length=6, variable_signal_peptide=10, include_cysteine_n_glycosylation: bool=False):
        if constant_modifications is None:
            constant_modifications = []
        if variable_modifications is None:
            variable_modifications = []

        self.constant_modifications = constant_modifications
        self.variable_modifications = variable_modifications
        self.min_length = min_length
        self.variable_signal_peptide = variable_signal_peptide
        self.peptide_permuter = PeptidePermuter(
            self.constant_modifications, self.variable_modifications)
        self.include_cysteine_n_glycosylation = include_cysteine_n_glycosylation

    def __reduce__(self):
        return self.__class__, (self.constant_modifications,
                                self.variable_modifications,
                                self.min_length,
                                self.variable_signal_peptide,
                                self.include_cysteine_n_glycosylation)

    def handle_protein(self, protein_obj: Protein, sites: Optional[Set[int]]=None):
        annot_sites = self.get_split_sites_from_annotations(protein_obj)
        if sites is None:
            try:
                accession = get_uniprot_accession(protein_obj.name)
                if accession:
                    try:
                        extra_sites, more_annots = self.get_split_sites_from_uniprot(accession)
                        # Trigger SQLAlchemy mutable update via assignment instead of in-place updates
                        protein_obj.annotations = list(protein_obj.annotations) + list(more_annots)
                        return self.split_protein(protein_obj, sorted(annot_sites | extra_sites))
                    except IOError:
                        return self.split_protein(protein_obj, annot_sites)
                else:
                    return self.split_protein(protein_obj, annot_sites)
            except XMLSyntaxError:
                return self.split_protein(protein_obj, annot_sites)
            except Exception as e:
                self.error(
                    ("An unhandled error occurred while retrieving"
                     " non-proteolytic cleavage sites"), e)
                return self.split_protein(protein_obj, annot_sites)
        else:
            if not isinstance(sites, set):
                sites = set(sites)
            try:
                return self.split_protein(protein_obj, sorted(sites | annot_sites))
            except IOError:
                return []

    def get_split_sites_from_annotations(self, protein: Protein) -> Set[int]:
        annots = protein.annotations
        split_sites = self._split_sites_from_annotations(annots)
        return split_sites

    def _split_sites_from_annotations(self, annots: annotation.AnnotationCollection) -> Set[int]:
        cleavable_annots = annots.cleavable()
        split_sites = set()
        for annot in cleavable_annots:
            split_sites.add(annot.start)
            split_sites.add(annot.end)
            if self.variable_signal_peptide and annot.feature_type == annotation.SignalPeptide.feature_type:
                for i in range(1, self.variable_signal_peptide + 1):
                    split_sites.add(annot.end + i)
        if 0 in split_sites:
            split_sites.remove(0)
        return split_sites

    def get_split_sites_from_features(self, record) -> Set[int]:
        if isinstance(record, uniprot.UniProtProtein):
            annots = annotation.from_uniprot(record)
        elif isinstance(record, annotation.AnnotationCollection):
            annots = record
        return self._split_sites_from_annotations(annots)

    def get_split_sites_from_uniprot(self, accession: str) -> Tuple[Set[int], annotation.AnnotationCollection]:
        try:
            record = uniprot.get(accession)
        except Exception as err:
            self.error(f"Failed to obtain Uniprot Record for {accession!r}: {err.__class__}:{str(err)}")
            return set()
        if record is None:
            self.error(f"Failed to obtain Uniprot Record for {accession!r}")
            return set()
        sites = self.get_split_sites_from_features(record)
        return sites, annotation.from_uniprot(record)

    def _make_split_expression(self, sites: List[int]) -> List:
        return [
            (Peptide.start_position < s) & (Peptide.end_position > s) for s in sites]

    def _permuted_peptides(self, sequence):
        return self.peptide_permuter(sequence)

    def split_protein(self, protein_obj: Protein, sites: Optional[List[int]]=None):
        if isinstance(sites, set):
            sites = sorted(sites)
        if sites is None:
            sites = []
        if not sites:
            return
        seen = set()
        sites_seen = set()
        peptides = protein_obj.peptides.all()
        peptide_intervals = IntervalTreeNode.build(map(PeptideInterval, peptides))
        for site in sites:
            overlap_region = peptide_intervals.contains_point(site - 1)
            spanned_intervals: IntervalTreeNode = IntervalTreeNode.build(overlap_region)
            # No spanned peptides. May be caused by regions of protein which digest to peptides
            # of unacceptable size.
            if spanned_intervals is None:
                continue
            lo = spanned_intervals.start
            hi = spanned_intervals.end
            # Get the set of all sites spanned by any peptide which spans the current query site
            spanned_sites = [s for s in sites if lo <= s <= hi]
            for i in range(1, len(spanned_sites) + 1):
                for split_sites in itertools.combinations(spanned_sites, i):
                    site_key = frozenset(split_sites)
                    if site_key in sites_seen:
                        continue
                    sites_seen.add(site_key)
                    spanning_peptides_query = spanned_intervals.contains_point(split_sites[0])
                    for site_j in split_sites[1:]:
                        spanning_peptides_query = [
                            sp for sp in spanning_peptides_query if site_j in sp
                        ]
                    spanning_peptides = []
                    for sp in spanning_peptides_query:
                        spanning_peptides.extend(sp)
                    for peptide in spanning_peptides:
                        peptide_start_position = peptide.start_position
                        adjusted_sites = [0] + [s - peptide_start_position for s in split_sites] + [
                            peptide.sequence_length]
                        for j in range(len(adjusted_sites) - 1):
                            # TODO: Cleavage sites may be off by one in the start. Revisit the index math here.
                            begin = adjusted_sites[j]
                            end = adjusted_sites[j + 1]
                            if end - begin < self.min_length:
                                continue
                            start_position = begin + peptide_start_position
                            end_position = end + peptide_start_position
                            if (start_position, end_position) in seen:
                                continue
                            else:
                                seen.add((start_position, end_position))
                            for modified_peptide, n_variable_modifications in self._permuted_peptides(
                                    peptide.base_peptide_sequence[begin:end]):

                                inst = Peptide(
                                    base_peptide_sequence=str(peptide.base_peptide_sequence[begin:end]),
                                    modified_peptide_sequence=str(modified_peptide),
                                    count_missed_cleavages=peptide.count_missed_cleavages,
                                    count_variable_modifications=n_variable_modifications,
                                    sequence_length=len(modified_peptide),
                                    start_position=start_position,
                                    end_position=end_position,
                                    calculated_mass=modified_peptide.mass,
                                    formula=formula(modified_peptide.total_composition()),
                                    protein_id=protein_obj.id)
                                inst.hypothesis_id = protein_obj.hypothesis_id
                                inst.peptide_score = 0
                                inst.peptide_score_type = 'null_score'
                                n_glycosites = n_glycan_sequon_sites(
                                    inst, protein_obj, include_cysteine=self.include_cysteine_n_glycosylation)
                                o_glycosites = o_glycan_sequon_sites(inst, protein_obj)
                                gag_glycosites = gag_sequon_sites(inst, protein_obj)
                                inst.count_glycosylation_sites = len(n_glycosites)
                                inst.n_glycosylation_sites = sorted(n_glycosites)
                                inst.o_glycosylation_sites = sorted(o_glycosites)
                                inst.gagylation_sites = sorted(gag_glycosites)
                                yield inst


class UniprotProteinAnnotator(TaskBase):
    hypothesis_builder: DatabaseBoundOperation
    protein_ids: List[int]
    constant_modifications: List[ModificationRule]
    variable_modifications: List[ModificationRule]
    variable_signal_peptide: int
    include_cysteine_n_glycosylation: bool

    def __init__(self, hypothesis_builder, protein_ids, constant_modifications,
                 variable_modifications, variable_signal_peptide=10,
                 include_cysteine_n_glycosylation: bool=False,
                 uniprot_source_file: Optional[str]=None):
        self.hypothesis_builder = hypothesis_builder
        self.protein_ids = protein_ids
        self.constant_modifications = constant_modifications
        self.variable_modifications = variable_modifications
        self.variable_signal_peptide = variable_signal_peptide
        self.session = hypothesis_builder.session
        self.include_cysteine_n_glycosylation = include_cysteine_n_glycosylation
        self.uniprot_source_file = uniprot_source_file

    def query(self, *args, **kwargs):
        return self.session.query(*args, **kwargs)

    def commit(self):
        return self.session.commit()

    def run(self):
        self.log("Begin Applying Protein Annotations")
        splitter = ProteinSplitter(self.constant_modifications, self.variable_modifications,
                                   variable_signal_peptide=self.variable_signal_peptide,
                                   include_cysteine_n_glycosylation=self.include_cysteine_n_glycosylation)
        i = 0
        j = 0
        protein_ids = self.protein_ids
        protein_names = [self.query(Protein.name).filter(
            Protein.id == protein_id).first()[0] for protein_id in protein_ids]
        name_to_id = {n: i for n, i in zip(protein_names, protein_ids)}
        if self.uniprot_source_file is not None:
            self.log(f"... Loading from local XML file {self.uniprot_source_file!r}")
            uniprot_queue = UniprotProteinXML(self.uniprot_source_file, protein_names)
        else:
            self.log("... Loading from UniProt web service")
            uniprot_queue = UniprotProteinDownloader(
                protein_names,
                n_threads=getattr(self.hypothesis_builder, "n_processes", cpu_count() // 2) * 4)
            uniprot_queue.start()
        n = len(protein_ids)
        interval = int(min((n * 0.1) + 1, 1000))
        acc = []
        seen = set()
        # TODO: This seems to periodically exit before finishing processing of all proteins?
        while True:
            try:
                protein_name, record = uniprot_queue.get()
                protein_id = name_to_id[protein_name]
                seen.add(protein_id)
                protein = self.query(Protein).get(protein_id)
                i += 1
                # Record not found in UniProt
                if isinstance(record, (Exception, str)) and not protein.annotations:
                    message = str(record)
                    # Many, many records that used to exist no longer do. We don't care if they are absent.
                    # if "Document is empty" not in message:
                    self.log(f"... Skipping {protein_name}: {message}")
                    continue
                if record is None and not protein.annotations:
                    self.log(f"... Skipping {protein_name}: no record found")
                    continue
                if i % interval == 0:
                    self.log(
                        f"... {i * 100. / n:0.3f}% Complete ({i}/{n}). {j} Peptides Produced.")

                if not isinstance(record, (uniprot.UniProtProtein, annotation.AnnotationCollection)):
                    sites = None
                else:
                    if isinstance(record, uniprot.UniProtProtein):
                        record = annotation.from_uniprot(record)
                    # Trigger SQLAlchemy mutable update via assignment instead of in-place updates
                    protein.annotations = list(protein.annotations) + list(record)
                    self.session.add(protein)
                    sites = splitter.get_split_sites_from_features(record)
                for peptide in splitter.handle_protein(protein, sites):
                    acc.append(peptide)
                    j += 1
                    if len(acc) > 100000:
                        self.log(
                            f"... {i * 100. / n:0.3f}% Complete ({i}/{n}). {j} Peptides Produced.")
                        self.session.bulk_save_objects(acc)
                        self.session.commit()
                        acc = []
            except Empty:
                uniprot_queue.join()
                if uniprot_queue.done_event.is_set():
                    break

        leftover_ids = set(protein_ids) - seen
        if leftover_ids:
            self.log(f"{len(leftover_ids)} IDs not covered by queue, sequentially querying UniProt")

        for protein_id in leftover_ids:
            i += 1
            protein = self.query(Protein).get(protein_id)
            if i % interval == 0:
                self.log(
                    f"... {i * 100. / n:0.3f}% Complete ({i}/{n}). {j} Peptides Produced.")
            for peptide in splitter.handle_protein(protein):
                acc.append(peptide)
                j += 1
                if len(acc) > 100000:
                    self.log(
                        f"... {i * 100. / n:0.3f}% Complete ({i}/{n}). {j} Peptides Produced.")
                    self.session.bulk_save_objects(acc)
                    self.session.commit()
                    acc = []
            self.session.add(protein)
        self.log(
            f"... {i * 100. / n:0.3f}% Complete ({i}/{n}). {j} Peptides Produced.")
        self.session.bulk_save_objects(acc)
        self.session.commit()
        acc = []



from collections import defaultdict
from contextlib import contextmanager
from typing import Dict, Optional, Tuple, Union
from ms_deisotope import DeconvolutedPeak

import numpy as np

from ms_deisotope.data_source import ProcessedRandomAccessScanSource
from ms_deisotope.output.common import LCMSMSQueryInterfaceMixin

from glycresoft.task import TaskBase
from glycresoft.chromatogram_tree import (
    ChromatogramForest,
    SimpleChromatogram,
    find_truncation_points,
    ChromatogramFilter)


class ChromatogramExtractor(TaskBase):
    peak_loader: Union[ProcessedRandomAccessScanSource, LCMSMSQueryInterfaceMixin]
    truncate: bool

    minimum_mass: float
    minimum_intensity: float

    grouping_tolerance: float
    min_points: int
    delta_rt: float

    peak_mapping: Optional[Dict[Tuple, DeconvolutedPeak]]
    chromatograms: Optional[ChromatogramForest]
    base_peak_chromatogram: Optional[SimpleChromatogram]
    total_ion_chromatogram: Optional[SimpleChromatogram]

    def __init__(self, peak_loader, truncate=False, minimum_mass=500, grouping_tolerance=1.5e-5,
                 minimum_intensity=250., min_points=3, delta_rt=0.25):
        self.peak_loader = peak_loader
        self.truncate = truncate
        self.minimum_mass = minimum_mass
        self.minimum_intensity = minimum_intensity
        self.grouping_tolerance = grouping_tolerance

        self.min_points = min_points
        self.delta_rt = delta_rt

        # self.accumulated = None
        # self.annotated_peaks = None
        self.peak_mapping = None

        self.chromatograms = None
        self.base_peak_chromatogram = None
        self.total_ion_chromatogram = None

    def get_scan_by_id(self, scan_id: str):
        return self.peak_loader.get_scan_by_id(scan_id)

    def get_scan_header_by_id(self, scan_id: str):
        return self.peak_loader.get_scan_header_by_id(scan_id)

    def get_index_information_by_scan_id(self, scan_id: str) -> dict:
        return self.peak_loader.get_index_information_by_scan_id(scan_id)

    def scan_id_to_rt(self, scan_id: str) -> float:
        return self.peak_loader.convert_scan_id_to_retention_time(scan_id)

    @contextmanager
    def toggle_peak_loading(self):
        if hasattr(self.peak_loader, 'toggle_peak_loading'):
            with self.peak_loader.toggle_peak_loading():
                yield
        else:
            yield

    def load_peaks(self):
        accumulated = self.peak_loader.ms1_peaks_above(min(500.0, self.minimum_mass), self.minimum_intensity)
        annotated_peaks = [x[:2] for x in accumulated]
        self.peak_mapping = {x[:2]: x[2] for x in accumulated}
        if len(accumulated) > 0:
            self.minimum_intensity = np.percentile([p[1].intensity for p in accumulated], 5)
        return annotated_peaks

    def aggregate_chromatograms(self, annotated_peaks):
        forest = ChromatogramForest([], self.grouping_tolerance, self.scan_id_to_rt)
        forest.aggregate_peaks(annotated_peaks, self.minimum_mass, self.minimum_intensity)
        chroma = list(forest)
        self.log("...... %d Chromatograms Extracted." % (len(chroma),))
        self.chromatograms = ChromatogramFilter.process(
            chroma, min_points=self.min_points, delta_rt=self.delta_rt)

    def summary_chromatograms(self, annotated_peaks):
        mapping = defaultdict(list)
        for scan_id, peak in annotated_peaks:
            mapping[scan_id].append(peak.intensity)
        bpc = SimpleChromatogram()
        tic = SimpleChromatogram()
        collection = sorted(mapping.items(), key=lambda b: self.scan_id_to_rt(b[0]))
        for scan_id, intensities in collection:
            rt = self.scan_id_to_rt(scan_id)
            bpc[rt] = max(intensities)
            tic[rt] = sum(intensities)
        self.base_peak_chromatogram = bpc
        self.total_ion_chromatogram = tic

    def run(self):
        self.log("... Begin Extracting Chromatograms")
        annotated_peaks = self.load_peaks()
        self.log("...... Aggregating Chromatograms")
        self.aggregate_chromatograms(annotated_peaks)
        self.summary_chromatograms(annotated_peaks)
        # Ensure chromatograms are wrapped and sorted.
        if self.truncate:
            self.chromatograms = ChromatogramFilter(
                self.truncate_chromatograms(self.chromatograms))
        else:
            self.chromatograms = ChromatogramFilter(self.chromatograms)
        return self.chromatograms

    def truncate_chromatograms(self, chromatograms):
        start, stop = find_truncation_points(*self.total_ion_chromatogram.as_arrays())
        out = []
        for c in chromatograms:
            if len(c) == 0:
                continue
            c.truncate_before(start)
            if len(c) == 0:
                continue
            c.truncate_after(stop)
            if len(c) == 0:
                continue
            out.append(c)
        return out

    def __iter__(self):
        if self.chromatograms is None:
            self.run()
        return iter(self.chromatograms)



cimport cython

from cpython.list cimport PyList_GetItem

cimport numpy as np
import numpy as np
np.import_array()
from numpy.math cimport isnan


cdef class ScoreCell(object):
    cdef:
        public double score
        public double value

    def __init__(self, score, value):
        self.score = score
        self.value = value

    def __getitem__(self, i):
        if i == 0:
            return self.score
        elif i == 1:
            return self.value
        else:
            raise ValueError(i)

    def __eq__(self, other):
        return abs(self.score - other.score) < 1e-12 and abs(self.value - other.value) < 1e-12

    def __ne__(self, other):
        return not self == other

    def __len__(self):
        return 2

    def __iter__(self):
        yield self.score
        yield self.value

    def __repr__(self):
        return "{self.__class__.__name__}({self.score}, {self.value})".format(self=self)

    def __reduce__(self):
        return self.__class__, (self.score, self.value)


cpdef __pyx_unpickle_ScoreCell(tp, checksum, state):
    return tp(*state)


cdef class NearestValueLookUp(object):
    '''A mapping-like object which simplifies
    finding the value of a pair whose key is nearest
    to a given query.

    .. note::
        Queries exceeding the maximum key will return
        the maximum key's value.
    '''
    cdef:
        public list items

    def __init__(self, items):
        if isinstance(items, dict):
            items = items.items()
        self.items = self._transform_items(items)

    def __eq__(self, other):
        return self.items == other.items

    def __ne__(self, other):
        return not self == other

    def _transform_items(self, items):
        return sorted([ScoreCell(*x) for x in items if not np.isnan(x[0])], key=lambda x: x[0])

    def __setstate__(self, state):
        if isinstance(state, tuple):
            self.items = self._transform_items(state[0])
        elif isinstance(state, dict):
            self.items = self._transform_items(state['items'])
        else:
            raise

    def __getstate__(self):
        return {
            "items": [tuple(i) for i in self.items]
        }

    cpdef max_key(self):
        cdef:
            Py_ssize_t n
            ScoreCell cell
        n = len(self.items)
        if n == 0:
            return 0
        cell = <ScoreCell>PyList_GetItem(self.items, n - 1)
        return cell.score

    cpdef Py_ssize_t _find_closest_item(self, double value):
        cdef:
            list array
            Py_ssize_t lo, hi, n, i, mid, best_index
            ScoreCell xc
            double error_tolerance, err, best_error, x

        array = self.items
        lo = 0
        hi = len(array)
        n = hi

        error_tolerance = 1e-3

        if isnan(value):
            return lo

        if lo == hi:
            return lo

        while hi - lo:
            i = (hi + lo) // 2
            xc = <ScoreCell>PyList_GetItem(array, i)
            x = xc.score
            err = x - value
            if abs(err) < error_tolerance:
                mid = i
                best_index = mid
                best_error = abs(err)
                i = mid - 1
                while i >= 0:
                    xc = <ScoreCell>PyList_GetItem(array, i)
                    x = xc.score
                    err = abs(x - value)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i -= 1
                i = mid + 1
                while i < n:
                    xc = <ScoreCell>PyList_GetItem(array, i)
                    x = xc.score
                    err = abs(x - value)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i += 1
                return best_index
            elif (hi - lo) == 1:
                mid = i
                best_index = mid
                best_error = abs(err)
                i = mid - 1
                while i >= 0:
                    xc = <ScoreCell>PyList_GetItem(array, i)
                    x = xc.score
                    err = abs(x - value)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i -= 1
                i = mid + 1
                while i < n:
                    xc = <ScoreCell>PyList_GetItem(array, i)
                    x = xc.score
                    err = abs(x - value)
                    if err < best_error:
                        best_error = err
                        best_index = i
                    elif err > error_tolerance:
                        break
                    i += 1
                return best_index
            elif x < value:
                lo = i
            elif x > value:
                hi = i

    cpdef get_pair(self, double key):
        return self.items[self._find_closest_item(key) + 1]

    def __len__(self):
        return len(self.items)

    def __repr__(self):
        return "{s.__class__.__name__}({size})".format(
            s=self, size=len(self))

    def __getitem__(self, key):
        return self._get_one(key)

    def _get_sequence(self, key):
        value = [self._get_one(k) for k in key]
        if isinstance(key, np.ndarray):
            value = np.array(value, dtype=float)
        return value

    cpdef _get_one(self, double key):
        ix = self._find_closest_item(key)
        if ix >= len(self):
            ix = len(self) - 1
        if ix < 0:
            ix = 0
        pair = self.items[ix]
        return pair.value


from glycresoft.serialize import (
    Peptide, Protein, DatabaseBoundOperation)

from .remove_duplicate_peptides import DeduplicatePeptides


class ProteomeBase(DatabaseBoundOperation):
    def __init__(self, connection, hypothesis_id, target_proteins=None,
                 constant_modifications=None, variable_modifications=None):
        if constant_modifications is None:
            constant_modifications = []
        if variable_modifications is None:
            variable_modifications = []
        DatabaseBoundOperation.__init__(self, connection)
        self.hypothesis_id = hypothesis_id
        self.target_proteins = target_proteins
        self.constant_modifications = constant_modifications
        self.variable_modifications = variable_modifications

    def retrieve_target_protein_ids(self):
        if len(self.target_proteins) == 0:
            return [
                i[0] for i in
                self.query(Protein.id).filter(
                    Protein.hypothesis_id == self.hypothesis_id).all()
            ]
        else:
            result = []
            for target in self.target_proteins:
                if isinstance(target, str):
                    match = self.query(Protein.id).filter(
                        Protein.name == target,
                        Protein.hypothesis_id == self.hypothesis_id).first()
                    if match:
                        result.append(match[0])
                    else:
                        self.log("Could not locate protein '%s'" % target)
                elif isinstance(target, int):
                    result.append(target)
            return result

    def get_target_proteins(self):
        ids = self.retrieve_target_protein_ids()
        return [self.session.query(Protein).get(i) for i in ids]

    def count_peptides(self):
        peptide_count = self.session.query(Peptide).filter(
            Peptide.hypothesis_id == self.hypothesis_id).count()
        return peptide_count

    def remove_duplicates(self):
        DeduplicatePeptides(self._original_connection, self.hypothesis_id).run()



from .evaluation import (
    SpectrumEvaluatorBase, LocalSpectrumEvaluator,
    SequentialIdentificationProcessor, SolutionHandler,
    SolutionPacker, MultiScoreSolutionHandler,
    MultiScoreSolutionPacker)

from .task import (
    StructureSpectrumSpecificationBuilder,
    TaskSourceBase, TaskDeque, TaskQueueFeeder)

from .utils import (
    SentinelToken, ProcessDispatcherState)

from .process import (
    IdentificationProcessDispatcher,
    SpectrumIdentificationWorkerBase)



import numbers as abc_numbers
from typing import Callable, Dict, Iterable, List, Optional, OrderedDict, DefaultDict as defaultdict, TYPE_CHECKING, Set, Tuple

import numpy as np

from glypy.structure.glycan_composition import FrozenMonosaccharideResidue, HashableGlycanComposition

from ..glycan_composition_filter import GlycanCompositionFilter
from ... import symbolic_expression

from .space import CompositionSpace, n_glycan_distance
from .rule import CompositionExpressionRule, CompositionRangeRule, CompositionRuleClassifier

if TYPE_CHECKING:
    from glycresoft.database.composition_network import CompositionGraph, CompositionGraphNode


class NeighborhoodCollection(object):
    neighborhoods: OrderedDict[str, CompositionRuleClassifier]

    def __init__(self, neighborhoods=None):
        if neighborhoods is None:
            neighborhoods = OrderedDict()
        self.neighborhoods = OrderedDict()
        if isinstance(neighborhoods, (dict)):
            self.neighborhoods = OrderedDict(neighborhoods)
        else:
            for item in neighborhoods:
                self.add(item)

    def add(self, classifier: CompositionRuleClassifier):
        self.neighborhoods[classifier.name] = classifier

    def remove(self, key: str):
        return self.neighborhoods.pop(key)

    def update(self, iterable: Iterable[CompositionRuleClassifier]):
        for case in iterable:
            self.add(case)

    def clear(self):
        self.neighborhoods.clear()

    def copy(self):
        return self.__class__(self.neighborhoods)

    clone = copy

    def __iter__(self):
        return iter(self.neighborhoods.values())

    def __repr__(self):
        return "NeighborhoodCollection(%s)" % (', '.join(self.neighborhoods.keys()))

    def __eq__(self, other):
        return list(self) == list(other)

    def __ne__(self, other):
        return not (self == other)

    def get_neighborhood(self, key: str) -> CompositionRuleClassifier:
        return self.neighborhoods[key]

    def __getitem__(self, key):
        try:
            return self.get_neighborhood(key)
        except KeyError:
            if isinstance(key, abc_numbers.Number):
                return tuple(self.neighborhoods.values())[key]
            else:
                raise

    def __len__(self):
        return len(self.neighborhoods)


def make_n_glycan_neighborhoods() -> NeighborhoodCollection:
    """Create broad N-glycan neighborhoods.

    This method is primarily designed for human-like N-glycan biosynthesis and does
    not deal well with abundant Gal-alpha Gal or LacDiNAc extensions.

    Returns
    -------
    NeighborhoodCollection
    """
    neighborhoods = NeighborhoodCollection()

    _neuraminic = "(%s)" % ' + '.join(map(str, (
        FrozenMonosaccharideResidue.from_iupac_lite("NeuAc"),
        FrozenMonosaccharideResidue.from_iupac_lite("NeuGc")
    )))
    _hexose = "(%s)" % ' + '.join(
        map(str, map(FrozenMonosaccharideResidue.from_iupac_lite, ['Hex', ])))
    _hexnac = "(%s)" % ' + '.join(
        map(str, map(FrozenMonosaccharideResidue.from_iupac_lite, ['HexNAc', ])))

    high_mannose = CompositionRangeRule(
        _hexose, 3, 12) & CompositionRangeRule(
        _hexnac, 2, 2) & CompositionRangeRule(
        _neuraminic, 0, 0)
    high_mannose.name = "high-mannose"
    neighborhoods.add(high_mannose)

    base_hexnac = 3
    base_neuac = 2
    for i, spec in enumerate(['hybrid', 'bi', 'tri', 'tetra', 'penta', "hexa", "hepta"]):
        if i == 0:
            rule = CompositionRangeRule(
                _hexnac, base_hexnac - 1, base_hexnac + 1
            ) & CompositionRangeRule(
                _neuraminic, 0, base_neuac) & CompositionRangeRule(
                _hexose, base_hexnac + i - 1,
                base_hexnac + i + 3)
            rule.name = spec
            neighborhoods.add(rule)
        else:
            sialo = CompositionRangeRule(
                _hexnac, base_hexnac + i - 1, base_hexnac + i + 1
            ) & CompositionRangeRule(
                _neuraminic, 1, base_neuac + i
            ) & CompositionRangeRule(
                _hexose, base_hexnac + i - 1,
                base_hexnac + i + 2)

            sialo.name = "%s-antennary" % spec
            asialo = CompositionRangeRule(
                _hexnac, base_hexnac + i - 1, base_hexnac + i + 1
            ) & CompositionRangeRule(
                _neuraminic, 0, 1 if i < 2 else 0
            ) & CompositionRangeRule(
                _hexose, base_hexnac + i - 1,
                base_hexnac + i + 2)

            asialo.name = "asialo-%s-antennary" % spec
            neighborhoods.add(sialo)
            neighborhoods.add(asialo)
    return neighborhoods


def make_mammalian_n_glycan_neighborhoods() -> NeighborhoodCollection:
    """Create broad N-glycan neighborhoods allowing for mammalian N-glycans.

    This method deals with N-glycosyltransferases not found in humans (specifically Gal-alpha Gal)
    which skew the boundaries between neighborhoods on the Hexose dimension. :func:`make_n_glycan_neighborhoods`
    already handles NeuGc trivially.

    Returns
    -------
    NeighborhoodCollection
    """
    neighborhoods = NeighborhoodCollection()

    _neuraminic = "(%s)" % ' + '.join(map(str, (
        FrozenMonosaccharideResidue.from_iupac_lite("NeuAc"),
        FrozenMonosaccharideResidue.from_iupac_lite("NeuGc")
    )))
    _terminal = _neuraminic + \
        " + max(%s - %%d, 0)" % FrozenMonosaccharideResidue.from_iupac_lite("Hex")
    _hexose = "(%s)" % ' + '.join(
        map(str, map(FrozenMonosaccharideResidue.from_iupac_lite, ['Hex', ])))
    _hexnac = "(%s)" % ' + '.join(
        map(str, map(FrozenMonosaccharideResidue.from_iupac_lite, ['HexNAc', ])))

    high_mannose = CompositionRangeRule(
        _hexose, 3, 12) & CompositionRangeRule(
        _hexnac, 2, 2) & CompositionRangeRule(
        _neuraminic, 0, 0)
    high_mannose.name = "high-mannose"
    neighborhoods.add(high_mannose)

    base_hexnac = 3
    base_terminal_groups = 2
    for i, spec in enumerate(['hybrid', 'bi', 'tri', 'tetra', 'penta', "hexa", "hepta"]):
        if spec == 'hybrid':
            rule = CompositionRangeRule(
                _hexnac, base_hexnac - 1, base_hexnac + 1
            ) & CompositionRangeRule(
                _neuraminic, 0, base_terminal_groups) & CompositionRangeRule(
                _hexose, base_hexnac + i - 1,
                base_hexnac + i + 3)
            rule.name = spec
            neighborhoods.add(rule)
        else:
            sialo = CompositionRangeRule(
                _hexnac, base_hexnac + i - 1, base_hexnac + i + 1
            ) & CompositionRangeRule(
                (_neuraminic), 1, base_terminal_groups + i
            ) & CompositionExpressionRule(
                "(Hex > %d) & (Hex < (%d - (NeuAc + NeuGc)))" % (base_hexnac + i - 2, base_hexnac + (2 * i) + 3))

            sialo.name = "%s-antennary" % spec
            asialo = CompositionRangeRule(
                _hexnac, base_hexnac + i - 1, base_hexnac + i + 1
            ) & CompositionRangeRule(
                _neuraminic, 0, 1 if i < 2 else 0
            ) & CompositionRangeRule(
                _hexose, base_hexnac + i - 1,
                base_hexnac + (2 * i) + 3)
            asialo.name = "asialo-%s-antennary" % spec
            neighborhoods.add(sialo)
            neighborhoods.add(asialo)
    return neighborhoods


def make_adjacency_neighborhoods(network: "CompositionGraph"):
    space = CompositionSpace([node.composition for node in network])

    rules = []
    for node in network:
        terms = []
        for monosaccharide in space.monosaccharides:
            terms.append(("abs({} - {})".format(
                monosaccharide, node.composition[monosaccharide])))
        expr = '(%s) < 2' % (' + '.join(terms),)
        expr_rule = CompositionExpressionRule(expr)
        rule = CompositionRuleClassifier(str(node.composition), [expr_rule])
        rules.append(rule)
    return rules


_n_glycan_neighborhoods = make_n_glycan_neighborhoods()


class NeighborhoodWalker(object):
    network: "CompositionGraph"
    neighborhoods: NeighborhoodCollection
    distance_fn: Callable[[HashableGlycanComposition, HashableGlycanComposition], Tuple[int, float]]

    symbols: symbolic_expression.SymbolSpace
    filter_space: GlycanCompositionFilter

    neighborhood_assignments: defaultdict["CompositionGraphNode", Set[str]]
    neighborhood_maps: defaultdict[str, List["CompositionGraphNode"]]

    def __init__(self, network, neighborhoods=None, assign=True, distance_fn=n_glycan_distance):
        if neighborhoods is None:
            if network.neighborhoods:
                neighborhoods = NeighborhoodCollection(network.neighborhoods)
            else:
                neighborhoods = NeighborhoodCollection(_n_glycan_neighborhoods)
        self.network = network
        self.neighborhood_assignments = defaultdict(set)
        self.neighborhoods = neighborhoods
        self.distance_fn = distance_fn
        self.filter_space = GlycanCompositionFilter(
            [self.normalize_composition(node.composition) for node in self.network])

        self.symbols = symbolic_expression.SymbolSpace(self.filter_space.monosaccharides)

        self.neighborhood_maps = defaultdict(list)

        if assign:
            self.assign()

    def normalize_composition(self, composition):
        return self.network.normalize_composition(composition)

    def _pack_maps(self):
        key_neighborhood_assignments = defaultdict(set)
        key_neighborhood_maps = defaultdict(list)

        for key, value in self.neighborhood_assignments.items():
            key_neighborhood_assignments[key.glycan_composition] = value
        for key, value in self.neighborhood_maps.items():
            key_neighborhood_maps[key.glycan_composition] = value
        return key_neighborhood_assignments, key_neighborhood_maps

    def _unpack_maps(self, packed_maps):
        (key_neighborhood_assignments, key_neighborhood_maps) = packed_maps

        for key, value in key_neighborhood_assignments.items():
            self.neighborhood_assignments[self.network[key.glycan_composition]] = value

        for key, value in key_neighborhood_maps.items():
            self.neighborhood_maps[self.network[key.glycan_composition]] = value

    def __getstate__(self):
        return self._pack_maps()

    def __setstate__(self, state):
        self._unpack_maps(state)

    def __reduce__(self):
        return self.__class__, (self.network, self.neighborhoods, False)

    def neighborhood_names(self):
        return [n.name for n in self.neighborhoods]

    def __getitem__(self, key):
        return self.neighborhood_assignments[key]

    def query_neighborhood(self, neighborhood: CompositionRuleClassifier):
        query = None
        filters = []
        for rule in neighborhood.rules:
            if not self.symbols.partially_defined(rule.symbols):
                continue

            filters.append(rule)
            try:
                low = rule.low
                high = rule.high
            except AttributeError:
                continue
            if low is None:
                low = 0
            if high is None:
                # No glycan will have more than 100 of a single residue
                # in practice.
                high = 100
            name = rule.symbols[0]
            if query is None:
                query = self.filter_space.query(name, low, high)
            else:
                query.add(name, low, high)
        if filters:
            query = filter(lambda x: all([f(x) for f in filters]), query)
        else:
            query = query.all()
        return query

    def assign(self):
        for neighborhood in self.neighborhoods:
            query = self.query_neighborhood(neighborhood)
            if query is None:
                raise ValueError("Query cannot be None! %r" % neighborhood)
            for composition in query:
                composition = self.normalize_composition(composition)
                if neighborhood(composition):
                    self.neighborhood_assignments[
                        self.network[composition]].add(neighborhood.name)
        for node in self.network:
            for neighborhood in self[node]:
                self.neighborhood_maps[neighborhood].append(node)

    def compute_belongingness(self, node: "CompositionGraphNode", neighborhood: CompositionRuleClassifier,
                              distance_cache: Optional[Dict]=None) -> float:
        count = 0
        total_weight = 0
        distance_fn = self.distance_fn
        for member in self.neighborhood_maps[neighborhood]:
            if distance_cache is None:
                distance, weight = distance_fn(node.glycan_composition, member.glycan_composition)
            else:
                # assumes that distance is symmetric
                key = (node.index, member.index) if node.index < member.index else (member.index, node.index)
                try:
                    distance, weight = distance_cache[key]
                except KeyError:
                    distance, weight = distance_fn(node.glycan_composition, member.glycan_composition)
                    distance_cache[key] = distance, weight

            if distance == 0:
                weight = 1.0
            total_weight += weight
            count += 1
        if count == 0:
            return 0
        return total_weight / count

    def build_belongingness_matrix(self) -> np.ndarray:
        neighborhood_count = len(self.neighborhoods)
        belongingness_matrix = np.zeros(
            (len(self.network), neighborhood_count))

        distance_cache = dict()

        for node in self.network:
            was_in = self.neighborhood_assignments[node]
            for i, neighborhood in enumerate(self.neighborhoods):
                if neighborhood.name in was_in:
                    belongingness_matrix[node.index, i] = self.compute_belongingness(
                        node, neighborhood.name, distance_cache=distance_cache)
        return belongingness_matrix



from typing import (Any, Callable, Collection, DefaultDict, Dict, Hashable,
                    List, Optional, Set, Tuple, NamedTuple, Type, Union,
                    TYPE_CHECKING)

from glycresoft.chromatogram_tree.mass_shift import MassShiftBase, Unmodified
from glycresoft.chromatogram_tree.relation_graph import (
    ChromatogramGraph, ChromatogramGraphEdge, ChromatogramGraphNode, TimeQuery)

from glycresoft.task import TaskBase

from ..spectrum_match import SpectrumMatch

from .base import Predicate, always, SolutionEntry, TargetType, parsimony_sort, KeyMassShift, KeyTargetMassShift, build_glycopeptide_key
from .revision import MS2RevisionValidator, SpectrumMatchBackFiller


if TYPE_CHECKING:
    from .chromatogram import TandemAnnotatedChromatogram


class MassShiftDeconvolutionGraphNode(ChromatogramGraphNode):
    chromatogram: 'TandemAnnotatedChromatogram'

    def __init__(self, chromatogram, index, edges=None):
        super(MassShiftDeconvolutionGraphNode, self).__init__(
            chromatogram, index, edges=edges)

    def most_representative_solutions(self, threshold_fn: Predicate = always, reject_shifted: bool = False, percentile_threshold: float = 1e-5) -> List['SolutionEntry']:
        """Find the most representative solutions, the (very nearly the same, hopefully) structures with
        the highest aggregated score across all MSn events assigned to this collection.

        Parameters
        ----------
        threshold_fn: Callable
            A function that filters out invalid solutions based on some criteria, e.g.
            not passing the FDR threshold.
        reject_shifted: bool
            Whether or not to omit any solution that was not mass-shifted. Defaults to False
        percentile_threshold : float, optional
            The difference between the worst and best percentile to be reported. Defaults to 1e-5.

        Returns
        -------
        list
            A list of solutions with approximately the greatest weight
        """
        return self.chromatogram.most_representative_solutions(
            threshold_fn=threshold_fn, reject_shifted=reject_shifted,
            percentile_threshold=percentile_threshold
        )

    def solutions_for(self, structure, threshold_fn: Predicate = always, reject_shifted: bool = False) -> List[SpectrumMatch]:
        return self.chromatogram.solutions_for(
            structure, threshold_fn=threshold_fn, reject_shifted=reject_shifted)

    @property
    def mass_shifts(self):
        return self.chromatogram.mass_shifts

    @property
    def tandem_solutions(self):
        return self.chromatogram.tandem_solutions

    @property
    def total_signal(self):
        return self.chromatogram.total_signal

    @property
    def entity(self):
        return self.chromatogram.entity

    def best_match_for(self, structure, threshold_fn: Predicate = always) -> SpectrumMatch:
        solutions = self.solutions_for(structure, threshold_fn=threshold_fn)
        if not solutions:
            raise KeyError(structure)
        best_score = -float('inf')
        best_match = None
        for sol in solutions:
            if best_score < sol.score:
                best_score = sol.score
                best_match = sol
        return best_match


class MassShiftDeconvolutionGraphEdge(ChromatogramGraphEdge):
    def __hash__(self):
        return hash(frozenset((self.node_a.index, self.node_b.index)))

    def __eq__(self, other):
        if frozenset((self.node_a.index, self.node_b.index)) == frozenset((other.node_a.index, other.node_b.index)):
            return self.transition == other.transition
        return False

    def __ne__(self, other):
        return not self == other


class MassShiftDeconvolutionGraph(ChromatogramGraph[MassShiftDeconvolutionGraphNode, MassShiftDeconvolutionGraphEdge]):
    node_cls = MassShiftDeconvolutionGraphNode
    edge_cls = MassShiftDeconvolutionGraphEdge

    mass_shifts: Set[MassShiftBase]

    def __init__(self, chromatograms: List['TandemAnnotatedChromatogram']):
        super(MassShiftDeconvolutionGraph, self).__init__(chromatograms)
        mass_shifts = set()
        for node in self:
            mass_shifts.update(node.mass_shifts)
        self.mass_shifts = mass_shifts

    @property
    def tandem_solutions(self):
        solution_sets = []
        for node in self.nodes:
            solution_sets.extend(node.tandem_solutions)
        return solution_sets

    def __iter__(self):
        return iter(self.nodes)

    def _construct_graph_nodes(self, chromatograms: List['TandemAnnotatedChromatogram']):
        nodes = []
        for i, chroma in enumerate(chromatograms):
            node = (self.node_cls(chroma, i))
            nodes.append(node)
            if node.chromatogram.composition:
                self.enqueue_seed(node)
                self.assignment_map[node.chromatogram.composition] = node
        return nodes

    def find_edges(self, node: MassShiftDeconvolutionGraphNode, query_width: float = 0.01, threshold_fn: Predicate = always,  **kwargs):
        query = TimeQuery(node.chromatogram, query_width)
        nodes: List[MassShiftDeconvolutionGraphNode] = self.rt_tree.overlaps(
            query.start, query.end)

        structure = node.chromatogram.structure

        for other in nodes:
            solutions = other.solutions_for(
                structure, threshold_fn=threshold_fn)
            if solutions:
                shifts_in_solutions = frozenset(
                    {m.mass_shift for m in solutions})
                self.edges.add(
                    self.edge_cls(node, other, (frozenset(node.mass_shifts), shifts_in_solutions)))

    def build(self, query_width: float = 0.01, threshold_fn: Predicate = always, **kwargs):
        for node in self.iterseeds():
            self.find_edges(node, query_width=query_width,
                            threshold_fn=threshold_fn, **kwargs)


class RepresenterDeconvolution(TaskBase):
    max_depth = 10
    threshold_fn: Predicate
    key_fn: Callable[[TargetType], Hashable]

    group: List[MassShiftDeconvolutionGraphNode]

    conflicted: DefaultDict[MassShiftDeconvolutionGraphNode, List[KeyTargetMassShift]]
    solved: DefaultDict[KeyTargetMassShift, List[MassShiftDeconvolutionGraphNode]]
    participants: DefaultDict[KeyTargetMassShift, List[MassShiftDeconvolutionGraphNode]]
    best_scores: DefaultDict[KeyMassShift, float]

    key_to_solutions: DefaultDict[KeyTargetMassShift, Set[TargetType]]

    spectrum_match_backfiller: Optional["SpectrumMatchBackFiller"]
    revision_validator: MS2RevisionValidator

    def __init__(self, group, threshold_fn=always, key_fn=build_glycopeptide_key,
                 spectrum_match_backfiller: Optional["SpectrumMatchBackFiller"]=None,
                 revision_validator: Optional[MS2RevisionValidator]=None):
        if revision_validator is None:
            revision_validator = MS2RevisionValidator(threshold_fn)
        self.group = group
        self.threshold_fn = threshold_fn
        self.key_fn = key_fn
        self.spectrum_match_backfiller = spectrum_match_backfiller
        self.revision_validator = revision_validator

        self.participants = None
        self.key_to_solutions = None
        self.conflicted = None
        self.solved = None
        self.best_scores = None

        self.build()

    def _simplify(self):
        # When there is an un-considered double shift, this could lead to
        # choosing that double-shift being called a trivial solution and
        # locking in a specific interpretation regardless of weight. So
        # this method won't be used automatically.

        # Search all members' assignment lists and see if any
        # can be easily assigned.
        for member, conflicts in list(self.conflicted.items()):
            # There is no conflict, trivially solved.
            if len(conflicts) == 1:
                self.solved[conflicts[0]].append(member)
                self.conflicted.pop(member)
            else:
                keys = {k for k, e, m in conflicts}
                # There are multiple structures, but they all share the same key, essentially
                # a positional isomer. Trivially solved, but need to use care when merging to avoid
                # combining the different isomers.
                if len(keys) == 1:
                    self.solved[conflicts[0]].append(member)
                    self.conflicted.pop(member)

    def build(self):
        '''Construct the initial graph, populating all the maps
        and then correcting for unsupported mass shifts.
        '''
        participants = DefaultDict(list)
        conflicted = DefaultDict(list)
        key_to_solutions = DefaultDict(set)
        best_scores = DefaultDict(float)

        for member in self.group:
            for part in member.most_representative_solutions(
                    threshold_fn=self.threshold_fn, reject_shifted=False, percentile_threshold=0.4):
                key = self.key_fn(part.solution)
                mass_shift = part.match.mass_shift
                score = part.best_score
                key_to_solutions[key].add(part.solution)
                participants[key, part.solution, mass_shift].append(member)
                best_scores[key, mass_shift] = max(score, best_scores[key, mass_shift])
                conflicted[member].append((key, part.solution, mass_shift))

            for part in member.most_representative_solutions(
                    threshold_fn=self.threshold_fn, reject_shifted=True, percentile_threshold=0.4):
                key = self.key_fn(part.solution)
                mass_shift = part.match.mass_shift
                score = part.best_score
                key_to_solutions[key].add(part.solution)
                best_scores[key, mass_shift] = max(score, best_scores[key, mass_shift])
                participants[key, part.solution, mass_shift].append(member)
                conflicted[member].append((key, part.solution, mass_shift))

        self.solved = DefaultDict(list)
        self.participants = participants
        self.conflicted = conflicted
        self.best_scores = best_scores
        self.key_to_solutions = key_to_solutions
        self.prune_unsupported_participants()

    def find_supported_participants(self) -> Dict[Tuple, bool]:
        has_unmodified = DefaultDict(bool)
        for (key, solution, mass_shift) in self.participants:
            if mass_shift == Unmodified:
                has_unmodified[key] = True
        return has_unmodified

    def prune_unsupported_participants(self, support_map: Dict[Tuple, bool] = None):
        '''Remove solutions that are not supported by an Unmodified
        condition except when there is *no alternative*.

        Uses :meth:`find_supported_participants` to label solutions.

        Returns
        -------
        pruned_conflicts : dict[list]
            A mapping from graph node to pruned solution
        restored : list
            A list of the solutions that would have been pruned
            but had to be restored to give a node support.
        '''

        if support_map is None:
            support_map = self.find_supported_participants()

        pruned_participants: Dict[KeyTargetMassShift, List[MassShiftDeconvolutionGraphNode]] = {}
        pruned_scores: Dict[Tuple[Hashable, MassShiftBase], float] = {}
        pruned_conflicts: DefaultDict[MassShiftDeconvolutionGraphNode,
                                      List[Tuple[Hashable, TargetType, MassShiftBase]]] = DefaultDict(list)
        pruned_solutions: List[TargetType] = []
        kept_solutions = []

        key: Hashable
        solution: TargetType
        mass_shift: MassShiftBase
        # Loop through the participants, removing links to unsupported solutions
        # and their metadata.
        for (key, solution, mass_shift), nodes in list(self.participants.items()):
            if not support_map[key]:
                q = (key, solution, mass_shift)
                try:
                    pruned_scores[key, mass_shift] = self.best_scores.pop((key, mass_shift))
                except KeyError:
                    pass
                pruned_participants[q] = self.participants.pop(q)
                pruned_solutions.append(solution)
                for node in nodes:
                    pruned_conflicts[node].append(q)
                    self.conflicted[node].remove(q)
            else:
                kept_solutions.append(solution)

        restored: List[MassShiftDeconvolutionGraphNode] = []
        # If all solutions for a given node have been removed, have to put them back.
        for node, options in list(self.conflicted.items()):
            if not options:
                alternatives = self.find_alternatives(
                    node, pruned_solutions)
                if alternatives:
                    for alt in alternatives:
                        solution = alt.target
                        mass_shift = alt.mass_shift
                        key = self.key_fn(solution)
                        q = key, solution, mass_shift

                        self.participants[q].append(node)
                        self.conflicted[node].append(q)
                        self.key_to_solutions[key].add(solution)
                        if self.threshold_fn(alt):
                            self.best_scores[key, mass_shift] = max(
                                alt.score, self.best_scores[key, mass_shift])
                        else:
                            self.best_scores[key, mass_shift] = 1e-3
                    continue

                restored.append(node)
                self.conflicted[node] = conflicts = pruned_conflicts.pop(node)
                for q in conflicts:
                    (key, solution, mass_shift) = q
                    try:
                        self.best_scores[key, mass_shift] = pruned_scores.pop(
                            (key, mass_shift))
                    except KeyError:
                        pass
                    self.participants[q].append(node)
        return pruned_conflicts, restored

    def find_alternatives(self, node: MassShiftDeconvolutionGraphNode, pruned_solutions, ratio_threshold=0.9):
        alternatives = {alt_solution for _,
                        alt_solution, _ in self.participants}
        ratios = []

        for solution in pruned_solutions:
            for sset in node.tandem_solutions:
                try:
                    sm = sset.solution_for(solution)
                    if not self.threshold_fn(sm):
                        continue
                    for alt in alternatives:
                        try:
                            sm2 = sset.solution_for(alt)
                            weight1 = sm.score
                            weight2 = sm2.score / weight1
                            if weight2 > ratio_threshold:
                                ratios.append((weight1, weight2, sm, sm2))
                        except KeyError:
                            continue
                except KeyError:
                    continue
        ratios.sort(key=lambda x: (x[0], x[1]), reverse=True)
        if not ratios:
            return []
        weight1, weight2, sm, sm2 = ratios[0]
        kept = []
        for pair in ratios:
            if weight1 - pair[0] > 1e-5:
                break
            if weight2 - pair[1] > 1e-5:
                break
            kept.append(sm2)
        return kept

    def resolve(self):
        '''For each conflicted node, try to assign it to a solution based upon
        the set of all solved nodes.

        Returns
        -------
        changes : int
            The number of nodes assigned during this execution.
        '''
        tree: DefaultDict[Hashable, Dict[MassShiftBase, List[MassShiftDeconvolutionGraphNode]]] = DefaultDict(dict)
        # Build a tree of key->mass_shift->members
        for (key, solution, mass_shift), members in self.solved.items():
            tree[key][mass_shift] = list(members)

        changes = 0

        # For each conflicted member, search through all putative explanations of the
        # chromatogram and see if another mass shift state has been identified
        # for that explanation's key. If so, and that explanation spans this
        # member, then add the putative explanation to the set of hits with weight
        # equal to that best score (sum across all supporting mass shifts for same key)
        # for that putative explanation.
        # Sort for determinism.
        for member, conflicts in sorted(
                self.conflicted.items(),
                key=lambda x: x[0].chromatogram.total_signal, reverse=True):
            hits: DefaultDict[KeyTargetMassShift, float] = DefaultDict(float)
            for key, solution, mass_shift in conflicts:
                # Other solved groups' keys to solved mass shifts
                shifts_to_solved = tree[key]
                for solved_mass_shift, solved_members in shifts_to_solved.items():
                    if any(m.overlaps(member) for m in solved_members):
                        hits[key, solution, mass_shift] += self.best_scores[key, solved_mass_shift]
            if not hits:
                continue
            # Select the solution with the greatest total weight.
            ordered_options = sorted(hits.items(), key=lambda x: x[1], reverse=True)
            (best_key, best_solution, best_mass_shift), score = ordered_options[0]

            # Add a new solution to the tracker, and update the tree
            self.solved[best_key, best_solution, best_mass_shift].append(member)
            tree[best_key][best_mass_shift] = self.solved[best_key, best_solution, best_mass_shift]

            self.conflicted.pop(member)
            changes += 1
        return changes

    def total_weight_for_keys(self, score_map=None):
        if score_map is None:
            score_map = self.best_scores
        acc = DefaultDict(list)
        for (key, mass_shift), score in score_map.items():
            acc[key].append(score)
        return {
            key: sum(val) for key, val in acc.items()
        }

    def nominate_key_mass_shift(self, score_map=None):
        '''Find the key with the greatest total score
        over all mass shifts and nominate it and its
        best mass shift.

        Returns
        -------
        key : tuple
            The key tuple for the best solution
        mass_shift : :class:`~.MassShiftBase`
            The best mass shift for the best solution
        '''
        if score_map is None:
            score_map = self.best_scores
        totals = self.total_weight_for_keys(score_map)
        if len(totals) == 0:
            return None, None
        best_key, _ = max(totals.items(), key=lambda x: x[1])

        best_mass_shift = None
        best_score = -float('inf')
        for (key, mass_shift), score in score_map.items():
            if key == best_key:
                if best_score < score:
                    best_score = score
                    best_mass_shift = mass_shift
        return best_key, best_mass_shift

    def find_starting_point(self):
        '''Find a starting point for deconvolution and mark it as solved.

        Returns
        -------
        key : tuple
            The key tuple for the best solution
        best_match : object
            The target that belongs to this key-mass shift pair.
        mass_shift : :class:`~.MassShiftBase`
            The best mass shift for the best solution
        '''
        nominated_key, nominated_mass_shift = self.nominate_key_mass_shift()
        if nominated_key is None:
            return None, None, None
        best_node = best_match = None
        assignable = []

        for member, conflicts in self.conflicted.items():
            for key, solution, mass_shift in conflicts:
                if (key == nominated_key) and (mass_shift == nominated_mass_shift):
                    try:
                        match = member.best_match_for(solution, threshold_fn=self.threshold_fn)
                        if match is None:
                            continue
                        assignable.append((member, match))
                    except KeyError:
                        continue
        if not assignable:
            self.log("Could not find a solution matching %r %r. Trying to drop the mass shift." % (nominated_key, nominated_mass_shift))
            for member, conflicts in self.conflicted.items():
                for key, solution, mass_shift in conflicts:
                    self.log(key)
                    if (key == nominated_key):
                        self.log("Key Match %r for %r" % (key, solution))
                        try:
                            match = member.best_match_for(solution, threshold_fn=self.threshold_fn)
                            self.log("Match: %r" % match)
                            if match is None:
                                self.log("Match is None, skipping")
                                continue
                            assignable.append((member, match))
                        except KeyError as err:
                            self.log("KeyError, skipping: %r" % err)
                            continue


        if assignable:
            best_node, best_match = max(assignable, key=lambda x: x[1].score)

        self.solved[nominated_key, best_match.target, nominated_mass_shift].append(best_node)
        self.conflicted.pop(best_node)
        return nominated_key, best_match, nominated_mass_shift

    def recurse(self, depth=0):
        subgroup = list(self.conflicted)
        subprob = self.__class__(subgroup, threshold_fn=self.threshold_fn, key_fn=self.key_fn)
        subprob.solve(depth=depth + 1)

        for key, val in subprob.solved.items():
            self.solved[key].extend(val)

        n = len(self.conflicted)
        k = len(subprob.conflicted)
        changed = n - k
        self.conflicted = subprob.conflicted
        return changed

    def default(self):
        '''Assign each chromatogram to its highest scoring identification
        individually, the solution graph could not be deconvolved.
        '''
        for member, conflicts in list(self.conflicted.items()):
            entry = member.most_representative_solutions(threshold_fn=self.threshold_fn)[0]
            key = self.key_fn(entry.solution)
            self.solved[key, entry.solution, entry.match.mass_shift].append(member)
            self.conflicted.pop(member)

    def solve(self, depth=0):
        '''Deconvolve the graph, removing conflicts and adding nodes to the
        solved set.

        If no nodes are solved initially, :meth:`find_starting_point` is used to select one.

        Solutions are found using :meth:`resolve` to build on starting points and
        :meth:`recurse`

        '''
        if depth > self.max_depth:
            self.default()
            return self
        if not self.solved:
            nominated_key, _, _ = self.find_starting_point()
            if nominated_key is None:
                self.log("No starting point found in %r, defaulting.", self.group)
                self.default()
                return self

        i = 0
        while self.conflicted:
            changed = self.resolve()
            if changed == 0 and self.conflicted:
                changed = self.recurse(depth)
            if changed == 0:
                break
            i += 1
            if i > 20:
                break
        return self

    def _representer_best_scores_percentiles(self, member: 'TandemAnnotatedChromatogram', solutions: Set[TargetType]) -> Tuple[Dict[TargetType, SpectrumMatch], Dict[TargetType, float], Dict[TargetType, float]]:
        entries: Dict[TargetType, List[SpectrumMatch]] = dict()
        for sol in solutions:
            entries[sol] = member.solutions_for(
                sol, threshold_fn=self.threshold_fn)
        scores: Dict[TargetType, float] = {}
        total = 1e-6
        best_matches: Dict[TargetType, SpectrumMatch] = {}
        for k, v in entries.items():
            best_match: Optional[SpectrumMatch] = None
            sk: float = 0
            for match in v:
                if best_match is None or best_match.score < match.score:
                    best_match = match
                sk += match.score
            best_matches[k] = best_match
            scores[k] = sk
            total += sk
        percentiles = {k: v / total for k, v in scores.items()}
        return best_matches, scores, percentiles

    def _representer_solution_entries(self, member: 'TandemAnnotatedChromatogram',
                                      best_matches: Dict[TargetType, SpectrumMatch],
                                      scores: Dict[TargetType, float],
                                      percentiles: Dict[TargetType, float]) -> List[SolutionEntry]:
        sols = []
        for k, best_match in best_matches.items():
            score = scores[k]
            if best_match is None:
                try:
                    best_match = member.best_match_for(k)
                except KeyError:
                    continue
                score = best_match.score
            sols.append(
                SolutionEntry(
                    k, score, percentiles[k], best_match.score, best_match))

        sols = parsimony_sort(sols)
        return sols

    def assign_representers(self, percentile_threshold=1e-5):
        '''Apply solutions to solved graph nodes, constructing new chromatograms with
        corrected mass shifts, entities, and representative_solution attributes and then
        merge within keys as appropriate.

        Returns
        -------
        merged : list[ChromatogramWrapper]
            The merged chromatograms
        '''
        # Note suggested to switch to list, but reason unclear.
        assigned: DefaultDict[Any, Set['TandemAnnotatedChromatogram']] = DefaultDict(set)

        invalidated_alternatives: DefaultDict[Any, Set[TargetType]] = DefaultDict(set)

        # For each (key, mass shift) pair and their members, compute the set of representers for
        # that key from all structures that as subsumed into it (repeated peptide and alternative
        # localization).
        # NOTE: Alternative localizations fall under the same key, so we have to re-do ranking
        # of solutions again here to make sure that the aggregate scores are available to separate
        # different localizations. Alternatively, may need to port in localization-separating key
        # logic.
        _mass_shift: MassShiftBase
        for (key, _solution, _mass_shift), members in self.solved.items():
            # We do not use _mass_shift here because it will be inferred from the best solution
            # for `key` on the chromatogram anyway, we only needed it keep the hash distinct
            #
            # We look up solutions instead of using _solution from the loop here because
            # shared peptides between proteins need to be tracked separately.
            solutions = self.key_to_solutions[key]
            for member in members:
                # Summarize each explainer
                best_matches, scores, percentiles = self._representer_best_scores_percentiles(member, solutions)

                # Pack each explainer into a SolutionEntry instance
                sols = self._representer_solution_entries(member, best_matches, scores, percentiles)
                if sols:
                    # This difference is not using the absolute value to allow for scenarios where
                    # a worse percentile is located at position 0 e.g. when hoisting via parsimony.
                    representers = [x for x in sols if (
                        sols[0].percentile - x.percentile) < percentile_threshold]
                else:
                    representers = []

                # Mass shifts will be applied when we assign the representer
                fresh = member.chromatogram.clone().drop_mass_shifts()

                fresh.assign_entity(representers[0])
                fresh.representative_solutions = representers

                assigned[key].add(fresh)
                if key != self.key_fn(member.chromatogram.entity):
                    invalidated_alternatives[key].add(member.chromatogram.entity)

        merged = []
        for key, members in assigned.items():
            members: List['TandemAnnotatedChromatogram'] = sorted(members, key=lambda x: x.start_time)
            invalidated_targets = invalidated_alternatives[key]
            retained: List['TandemAnnotatedChromatogram'] = []

            member_targets: Set[TargetType] = set()
            for chrom in members:
                member_targets.add(chrom.entity)

            member_targets -= invalidated_targets

            can_merge: List['TandemAnnotatedChromatogram'] = []

            for member in members:
                if self.revision_validator.has_valid_matches(member, member_targets):
                    can_merge.append(member)
                else:
                    retained_solutions = member.most_representative_solutions(self.threshold_fn)
                    member.representative_solutions = retained_solutions
                    member.assign_entity(retained_solutions[0])
                    retained.append(member)

            sink = None
            if can_merge:
                sink = can_merge[0]
                did_update_sink_id = False
                if sink.entity not in member_targets:
                    self.error(f"... While deconvolving mass shifts, sink node {sink}'s label not in valid list {member_targets}")
                    options = sorted(
                        filter(
                            lambda x: x.solution in member_targets,
                            sink.compute_representative_weights(
                                self.threshold_fn)
                        ),
                        key=lambda x: x.score,
                        reverse=True
                    )
                    if options:
                        self.error(f"... Failed to re-assign {sink} to a valid target")
                        sink.assign_entity(options[0])
                    # May not be exhaustive enough, search for "best" case in `can_merge` matching
                    # to a `member_targets` before assigning `sink` role to a chromatogram?
                    did_update_sink_id = True

                for member in can_merge[1:]:
                    # Might this obscure the "best localization"? Perhaps, but localization is already getting
                    # mowed down at the aggregate level.
                    if self.revision_validator.can_rewrite_best_matches(member, sink.entity):
                        self.revision_validator.do_rewrite_best_matches(member, sink.entity, invalidated_targets)
                        sink.merge_in_place(member)
                    else:
                        retained_solutions = member.most_representative_solutions(self.threshold_fn)
                        member.representative_solutions = retained_solutions
                        member.assign_entity(retained_solutions[0])
                        retained.append(member)

                if did_update_sink_id:
                    options = sorted(
                        filter(
                            lambda x: x.solution in member_targets,
                            sink.compute_representative_weights(
                                self.threshold_fn)
                        ),
                        key=lambda x: x.score,
                        reverse=True
                    )
                    if options:
                        self.error(f"... Failed to re-assign {sink} to a valid target after merging!")
                        sink.assign_entity(options[0])

            if sink is not None:
                merged.append(sink)
            if retained:
                merged.extend(retained)
        return merged



# cython: embedsignature=True
from collections import defaultdict

cimport cython

from cpython.object cimport PyObject
from cpython.float cimport PyFloat_AsDouble
from cpython.int cimport PyInt_AsLong, PyInt_FromLong
from cpython.list cimport PyList_GET_SIZE, PyList_GET_ITEM, PyList_Append, PyList_Size, PyList_GetItem
from cpython.tuple cimport PyTuple_GET_SIZE, PyTuple_GET_ITEM, PyTuple_Size, PyTuple_GetItem
from cpython.dict cimport PyDict_GetItem, PyDict_SetItem, PyDict_Items, PyDict_Values
from cpython.set cimport PySet_Add, PySet_GET_SIZE

from libc cimport math

import numpy as np
cimport numpy as np
from numpy cimport npy_uint16 as uint16_t, npy_int32 as int32_t

from ms_deisotope._c.peak_set cimport DeconvolutedPeak, DeconvolutedPeakSet
from glycresoft._c.chromatogram_tree.mass_shift cimport MassShiftBase

np.import_array()

import warnings

from glypy.composition.ccomposition cimport CComposition
from glycopeptidepy._c.structure.fragmentation_strategy.glycan cimport StubGlycopeptideStrategy, GlycanCompositionFragment
from glycopeptidepy._c.count_table cimport CountTable
from glycresoft.chromatogram_tree import Unmodified as _Unmodified
from glycresoft.serialize import GlycanTypes

from glypy.structure.glycan_composition import HashableGlycanComposition as _HashableGlycanComposition

cdef object HashableGlycanComposition = _HashableGlycanComposition
cdef MassShiftBase Unmodified = _Unmodified
cdef object GlycanTypes_n_glycan, GlycanTypes_o_glycan, GlycanTypes_gag_linker


GlycanTypes_n_glycan = GlycanTypes.n_glycan
GlycanTypes_o_glycan = GlycanTypes.o_glycan
GlycanTypes_gag_linker = GlycanTypes.gag_linker


cdef class GlycanCombinationRecordBase(object):
    cdef:
        public size_t id
        public double dehydrated_mass
        public object composition
        public uint16_t size
        public uint16_t internal_size_approximation
        public uint16_t side_group_count
        public uint16_t count
        public list glycan_types
        public dict _fragment_cache
        public Py_hash_t _hash
        public dict fragment_set_properties

    cpdef bint is_n_glycan(self):
        return GlycanTypes_n_glycan in self.glycan_types

    cpdef bint is_o_glycan(self):
        return GlycanTypes_o_glycan in self.glycan_types

    cpdef bint is_gag_linker(self):
        return GlycanTypes_gag_linker in self.glycan_types

    @cython.cdivision(True)
    cpdef double normalized_size(self):
        cdef:
            double k
            double d
            uint16_t n
        n = self.internal_size_approximation
        if n == 0:
            return 0.0
        if self.side_group_count > 0:
            k = 1.0
        else:
            k = 2.0
        d = max(n * math.log(n) / k, n)
        return d

    cpdef list get_n_glycan_fragments(self):
        cdef:
            PyObject* ptemp
            list result, fragment_structs
        ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_n_glycan)
        if ptemp == NULL:
            strategy = StubGlycopeptideStrategy(None, extended=True)
            shifts = strategy.n_glycan_composition_fragments(self.composition, 1, 0)
            fragment_structs = []
            for shift in shifts:
                if shift["key"]['HexNAc'] <= 2 and shift["key"]["Hex"] <= 3:
                    is_core = True
                else:
                    is_core = False
                fragment_structs.append(
                    CoarseStubGlycopeptideFragment(
                        shift['key'], shift['mass'], is_core))
            result = sorted(set(fragment_structs))
            PyDict_SetItem(self._fragment_cache, GlycanTypes_n_glycan, result)
            return result
        else:
            result = <list>ptemp
            return result

    cpdef list get_o_glycan_fragments(self):
        cdef:
            PyObject* ptemp
            list result, fragment_structs
        ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_o_glycan)
        if ptemp == NULL:
            strategy = StubGlycopeptideStrategy(None, extended=True)
            shifts = strategy.o_glycan_composition_fragments(self.composition, 1, 0)
            fragment_structs = []
            for shift in shifts:
                fragment_structs.append(
                    CoarseStubGlycopeptideFragment(
                        shift['key'], shift['mass'], True))
            result = sorted(set(fragment_structs))
            PyDict_SetItem(self._fragment_cache, GlycanTypes_o_glycan, result)
            return result
        else:
            result = <list>ptemp
            return result

    cpdef list get_gag_linker_glycan_fragments(self):
        cdef:
            PyObject* ptemp
            list result, fragment_structs
        ptemp = PyDict_GetItem(self._fragment_cache, GlycanTypes_gag_linker)
        if ptemp == NULL:
            strategy = StubGlycopeptideStrategy(None, extended=True)
            shifts = strategy.gag_linker_composition_fragments(self.composition, 1, 0)
            fragment_structs = []
            for shift in shifts:
                fragment_structs.append(
                    CoarseStubGlycopeptideFragment(
                        shift['key'], shift['mass'], True))
            result = sorted(set(fragment_structs))
            PyDict_SetItem(self._fragment_cache, GlycanTypes_gag_linker, result)
            return result
        else:
            result = <list>ptemp
            return result

    cpdef clear(self):
        self._fragment_cache.clear()


cdef class CoarseStubGlycopeptideFragment(object):
    cdef:
        public object key
        public double mass
        public bint is_core

    def __init__(self, key, mass, is_core):
        self.key = key
        self.mass = mass
        self.is_core = is_core

    def __reduce__(self):
        return self.__class__, (self.key, self.mass, self.is_core)

    def __repr__(self):
        return "%s(%s, %f, %r)" % (
            self.__class__.__name__,
            self.key, self.mass, self.is_core
        )

    def __eq__(self, other):
        try:
            return self.key == other.key and self.is_core == other.is_core
        except AttributeError:
            return self.key == other

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(int(self.mass))

    def __lt__(self, other):
        return self.mass < other.mass

    def __gt__(self, other):
        return self.mass > other.mass


@cython.freelist(100000)
@cython.final
cdef class CoarseStubGlycopeptideMatch(object):
    cdef:
        public object key
        public double mass
        public double shift_mass
        public tuple peaks_matched

    @staticmethod
    cdef CoarseStubGlycopeptideMatch _create(object key, double mass, double shift_mass, tuple peaks_matched):
        cdef CoarseStubGlycopeptideMatch instance
        instance = CoarseStubGlycopeptideMatch.__new__(CoarseStubGlycopeptideMatch)
        instance.key = key
        instance.mass = mass
        instance.shift_mass = shift_mass
        instance.peaks_matched = peaks_matched
        return instance

    def __init__(self, key, mass, shift_mass, peaks_matched):
        self.key = key
        self.mass = mass
        self.shift_mass = shift_mass
        self.peaks_matched = peaks_matched

    def __reduce__(self):
        return self.__class__, (self.key, self.mass, self.shift_mass, self.peaks_matched)

    def __repr__(self):
        return "%s(%s, %f, %f, %r)" % (
            self.__class__.__name__,
            self.key, self.mass, self.shift_mass, self.peaks_matched
        )


@cython.freelist(100000)
@cython.final
cdef class CoarseGlycanMatch(object):
    cdef:
        public list fragment_matches
        public double n_matched
        public double n_theoretical
        public double core_matched
        public double core_theoretical

    @staticmethod
    cdef CoarseGlycanMatch _create(list fragment_matches, double n_matched, double n_theoretical, double core_matched, double core_theoretical):
        cdef CoarseGlycanMatch inst = CoarseGlycanMatch.__new__(CoarseGlycanMatch)
        inst.fragment_matches = fragment_matches
        inst.n_matched = n_matched
        inst.n_theoretical = n_theoretical
        inst.core_matched = core_matched
        inst.core_theoretical = core_theoretical
        return inst

    def __init__(self, fragment_matches, n_matched, n_theoretical, core_matched, core_theoretical):
        self.fragment_matches = list(fragment_matches)
        self.n_matched = n_matched
        self.n_theoretical = n_theoretical
        self.core_matched = core_matched
        self.core_theoretical = core_theoretical

    def __iter__(self):
        yield self.fragment_matches
        yield self.n_matched
        yield self.n_theoretical
        yield self.core_matched
        yield self.core_theoretical

    cpdef double estimate_peptide_mass(self):
        cdef:
            double weighted_mass_acc
            double weight_acc
            double fmass

            size_t i_fmatch, n_fmatch, i_peak, n_peak
            CoarseStubGlycopeptideMatch fmatch
            DeconvolutedPeak peak
        weighted_mass_acc = 0
        weight_acc = 0
        n_fmatch = PyList_GET_SIZE(self.fragment_matches)
        for i_fmatch in range(n_fmatch):
            fmatch = <CoarseStubGlycopeptideMatch>PyList_GET_ITEM(self.fragment_matches, i_fmatch)
            fmass = fmatch.shift_mass
            n_peak = PyTuple_GET_SIZE(fmatch.peaks_matched)
            for i_peak in range(n_peak):
                peak = <DeconvolutedPeak>PyTuple_GET_ITEM(fmatch.peaks_matched, i_peak)
                weighted_mass_acc += (peak.neutral_mass - fmass) * peak.intensity
                weight_acc += peak.intensity
        if weight_acc == 0:
            return -1
        return weighted_mass_acc / weight_acc

    def __repr__(self):
        template = (
            "{self.__class__.__name__}({self.n_matched}, {self.n_theoretical}, "
            "{self.core_matched}, {self.core_theoretical})")
        return template.format(self=self)


cdef class GlycanCoarseScorerBase(object):
    cdef:
        public double product_error_tolerance
        public double fragment_weight
        public double core_weight
        public bint recalibrate_peptide_mass

    def __init__(self, product_error_tolerance=1e-5, fragment_weight=0.56, core_weight=0.42, recalibrate_peptide_mass=False):
        self.product_error_tolerance = product_error_tolerance
        self.fragment_weight = fragment_weight
        self.core_weight = core_weight
        self.recalibrate_peptide_mass = recalibrate_peptide_mass

    @cython.boundscheck(False)
    cpdef CoarseGlycanMatch _match_fragments(self, object scan, double peptide_mass, list shifts, double mass_shift_tandem_mass=0.0):
        cdef:
            list fragment_matches
            double core_matched, core_theoretical, product_error_tolerance
            double n_frags_matched, approximate_size_normalizer
            DeconvolutedPeakSet peak_set
            size_t i, n
            bint has_tandem_shift
            tuple hits
            CoarseStubGlycopeptideFragment shift

        product_error_tolerance = (self.product_error_tolerance)
        fragment_matches = []

        core_matched = 0.0
        core_theoretical = 0.0
        n_frags_matched = 0.0
        has_tandem_shift = abs(mass_shift_tandem_mass) > 0
        peak_set = scan.deconvoluted_peak_set

        n = PyList_GET_SIZE(shifts)

        for i in range(n):
            shift = <CoarseStubGlycopeptideFragment>PyList_GET_ITEM(shifts, i)
            if shift.is_core:
                is_core = True
                core_theoretical += 1
            target_mass = shift.mass + peptide_mass
            hits = peak_set.all_peaks_for(target_mass, product_error_tolerance)
            if hits:
                if is_core:
                    core_matched += 1
                fragment_matches.append(
                    CoarseStubGlycopeptideMatch._create(shift.key, target_mass, shift.mass, hits))
                n_frags_matched += 1
            if has_tandem_shift:
                shifted_mass = target_mass + mass_shift_tandem_mass
                hits = peak_set.all_peaks_for(
                    shifted_mass, product_error_tolerance)
                if hits:
                    if is_core:
                        core_matched += 1
                    fragment_matches.append(
                        CoarseStubGlycopeptideMatch._create(
                            shift.key, shifted_mass, shift.mass + mass_shift_tandem_mass, hits))
                    n_frags_matched += 1

        return CoarseGlycanMatch._create(
            fragment_matches, n_frags_matched, float(n), core_matched, core_theoretical)

    @cython.cdivision(True)
    cpdef double _calculate_score(self, CoarseGlycanMatch glycan_match):
        cdef:
            size_t i, j, n, m
            CoarseStubGlycopeptideMatch matched_fragment
            double score, ratio_fragments, ratio_core, coverage
            DeconvolutedPeak peak
            tuple matches
            list matched_fragments

        ratio_fragments = (<double>glycan_match.n_matched) / glycan_match.n_theoretical
        ratio_core = <double>glycan_match.core_matched / glycan_match.core_theoretical
        coverage = (ratio_fragments ** (self.fragment_weight)) * (ratio_core ** (self.core_weight))
        score = 0
        matched_fragments = glycan_match.fragment_matches
        n = PyList_GET_SIZE(matched_fragments)
        for i in range(n):
            matched_fragment = <CoarseStubGlycopeptideMatch>PyList_GET_ITEM(matched_fragments, i)
            m = PyTuple_GET_SIZE(matched_fragment.peaks_matched)
            for j in range(m):
                peak = <DeconvolutedPeak>PyTuple_GET_ITEM(matched_fragment.peaks_matched, j)
                score += math.log(peak.intensity) * (1 - (
                    math.fabs(peak.neutral_mass - matched_fragment.mass) / matched_fragment.mass) ** 4)
        score *= coverage
        return score

    cpdef CoarseGlycanMatch _n_glycan_match_stubs(self, scan, double peptide_mass,
                                                  GlycanCombinationRecordBase glycan_combination,
                                                  double mass_shift_tandem_mass=0.0):
        shifts = glycan_combination.get_n_glycan_fragments()
        return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)

    cpdef CoarseGlycanMatch _o_glycan_match_stubs(self, scan, double peptide_mass,
                                                  GlycanCombinationRecordBase glycan_combination,
                                                  double mass_shift_tandem_mass=0.0):
        shifts = glycan_combination.get_o_glycan_fragments()
        return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)

    cpdef CoarseGlycanMatch _gag_match_stubs(self, scan, double peptide_mass,
                                             GlycanCombinationRecordBase glycan_combination,
                                             double mass_shift_tandem_mass=0.0):
        shifts = glycan_combination.get_gag_linker_glycan_fragments()
        return self._match_fragments(scan, peptide_mass, shifts, mass_shift_tandem_mass)

    cdef CoarseGlycanMatch _get_n_glycan_coarse_score(self, object scan,
                                                      GlycanCombinationRecordBase glycan_combination,
                                                      MassShiftBase mass_shift, double peptide_mass,
                                                      double* scoreout):
        if peptide_mass < 0:
            scoreout[0] = -1e6
            return None
        glycan_match = self._n_glycan_match_stubs(
            scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
        score = self._calculate_score(glycan_match)
        scoreout[0] = score
        return glycan_match

    cdef CoarseGlycanMatch _get_o_glycan_coarse_score(self, object scan,
                                                      GlycanCombinationRecordBase glycan_combination,
                                                      MassShiftBase mass_shift, double peptide_mass,
                                                      double* scoreout):
        if peptide_mass < 0:
            scoreout[0] = -1e6
            return None
        glycan_match = self._o_glycan_match_stubs(
            scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
        score = self._calculate_score(glycan_match)
        scoreout[0] = score
        return glycan_match

    cdef CoarseGlycanMatch _get_gag_coarse_score(self, object scan,
                                                 GlycanCombinationRecordBase glycan_combination,
                                                 MassShiftBase mass_shift, double peptide_mass,
                                                 double* scoreout):
        if peptide_mass < 0:
            scoreout[0] = -1e6
            return None
        glycan_match = self._gag_match_stubs(
            scan, peptide_mass, glycan_combination, mass_shift_tandem_mass=mass_shift.tandem_mass)
        score = self._calculate_score(glycan_match)
        scoreout[0] = score
        return glycan_match


cdef bint isclose(double a, double b, double rtol=1e-05, double atol=1e-08):
    return abs(a - b) <= atol + rtol * abs(b)


@cython.freelist(1000)
cdef class GlycanMatchResult(object):
    cdef:
        public double peptide_mass
        public double score
        public CoarseGlycanMatch match
        public size_t glycan_size
        public dict glycan_types
        public double recalibrated_peptide_mass


    @staticmethod
    cdef GlycanMatchResult _create(double peptide_mass, double score, CoarseGlycanMatch match, size_t glycan_size,
                                   dict glycan_types, double recalibrated_peptide_mass):
        cdef:
            GlycanMatchResult self

        self = GlycanMatchResult.__new__(GlycanMatchResult)
        self.peptide_mass = peptide_mass
        self.score = score
        self.match = match
        self.glycan_size = glycan_size
        self.glycan_types = glycan_types
        self.recalibrated_peptide_mass = recalibrated_peptide_mass
        return self

    def __init__(self, peptide_mass, score, match, glycan_size, glycan_types, recalibrated_peptide_mass):
        self.peptide_mass = peptide_mass
        self.score = score
        self.match = match
        self.glycan_size = glycan_size
        self.glycan_types = glycan_types
        self.recalibrated_peptide_mass = recalibrated_peptide_mass

    @property
    def fragment_match_count(self):
        match = self.match
        if match is None:
            return 0
        return match.n_matched

    def __repr__(self):
        template = (
            "{self.__class__.__name__}({self.peptide_mass}, {self.score},"
            " {self.match}, {self.glycan_size}, {self.glycan_types}, {self.recalibrated_peptide_mass})")
        return template.format(self=self)


@cython.freelist(1000)
@cython.final
cdef class ScoreCoarseGlycanMatchPair(object):
    cdef:
        public double score
        public CoarseGlycanMatch match

    @staticmethod
    cdef ScoreCoarseGlycanMatchPair _create(double score, CoarseGlycanMatch match):
        cdef ScoreCoarseGlycanMatchPair self = ScoreCoarseGlycanMatchPair.__new__(ScoreCoarseGlycanMatchPair)
        self.score = score
        self.match = match
        return self

    def __init__(self, score, match):
        self.score = score
        self.match = match

    def __getitem__(self, i):
        if i == 0:
            return self.score
        elif i == 1:
            return self.match
        else:
            raise IndexError(i)

    def __len__(self):
        return 2

    def __repr__(self):
        return "(%f, %r)" % (self.score, self.match)


cpdef double score_getter(object x):
    return (<GlycanMatchResult>x).score


@cython.boundscheck(False)
@cython.binding(True)
cpdef GlycanFilteringPeptideMassEstimator_match(GlycanCoarseScorerBase self, scan, MassShiftBase mass_shift=Unmodified, query_mass=None):
    cdef:
        list output, glycan_combination_db
        double intact_mass, threshold_mass, last_peptide_mass, peptide_mass
        GlycanCombinationRecordBase glycan_combination
        dict type_to_score
        double best_score, score
        size_t i, n
        tuple match_stat
        CoarseGlycanMatch best_match, match

    output = []
    if query_mass is None:
        intact_mass = scan.precursor_information.neutral_mass
    else:
        intact_mass = query_mass
    threshold_mass = (intact_mass + 1) - self.minimum_peptide_mass
    last_peptide_mass = 0
    glycan_combination_db = self.glycan_combination_db
    n = PyList_Size(glycan_combination_db)
    for i in range(n):
        glycan_combination = <GlycanCombinationRecordBase>PyList_GET_ITEM(glycan_combination_db, i)
        # Stop searching when the peptide mass would be below the minimum peptide mass
        if threshold_mass < glycan_combination.dehydrated_mass:
            break
        peptide_mass = (
            intact_mass - glycan_combination.dehydrated_mass
        ) - mass_shift.mass
        best_score = 0
        best_match = None
        type_to_score = {}
        if glycan_combination.is_n_glycan():
            score = 0
            match = self._get_n_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
            type_to_score[GlycanTypes_n_glycan] = ScoreCoarseGlycanMatchPair._create(score, match)
            if score > best_score:
                best_score = score
                best_match = match
        if glycan_combination.is_o_glycan():
            score = 0
            match = self._get_o_glycan_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
            type_to_score[GlycanTypes_o_glycan] = ScoreCoarseGlycanMatchPair._create(score, match)
            if score > best_score:
                best_score = score
                best_match = match
        if glycan_combination.is_gag_linker():
            match = self._get_gag_coarse_score(scan, glycan_combination, mass_shift, peptide_mass, &score)
            type_to_score[GlycanTypes_gag_linker] = ScoreCoarseGlycanMatchPair._create(score, match)
            if score > best_score:
                best_score = score
                best_match = match

        if best_match is not None and self.recalibrate_peptide_mass:
            recalibrated_peptide_mass = best_match.estimate_peptide_mass()
            if recalibrated_peptide_mass > 0:
                if abs(recalibrated_peptide_mass - peptide_mass) > 0.5:
                    warnings.warn("Re-estimated peptide mass error is large: %f vs %f" % (
                        peptide_mass, recalibrated_peptide_mass))
        else:
            recalibrated_peptide_mass = 0
        result = GlycanMatchResult._create(
            peptide_mass,
            best_score, best_match, glycan_combination.size, type_to_score, recalibrated_peptide_mass)
        output.append(result)
    output = sorted(output, key=score_getter, reverse=1)
    return output


cdef class IndexGlycanCompositionFragment(GlycanCompositionFragment):
    cdef:
        public int index
        public str name

    def __init__(self, mass, composition, key, is_extended=False):
        self.mass = mass
        self.composition = composition
        self.key = key
        self.is_extended = is_extended
        self._hash_key = -1
        self.index = -1
        self.name = None

    @staticmethod
    cdef IndexGlycanCompositionFragment _create(double mass, CComposition composition, CountTable key, bint is_extended):
        cdef:
            IndexGlycanCompositionFragment self
        self = IndexGlycanCompositionFragment.__new__(IndexGlycanCompositionFragment)
        self.mass = mass
        self.composition = composition
        self.key = key
        self.is_extended = is_extended
        self._hash_key = -1
        self.index = -1
        self.name = None
        return self

    cpdef GlycanCompositionFragment copy(self):
        cdef:
            IndexGlycanCompositionFragment inst
        inst = IndexGlycanCompositionFragment._create(
            self.mass,
            self.composition.clone() if self.composition is not None else None, self.key.copy(),
            self.is_extended)
        inst._hash_key = self._hash_key
        inst.index = self.index
        inst.name = self.name
        return inst

    def __reduce__(self):
        return self.__class__, (self.mass, self.composition, self.key, self.is_extended), self.__getstate__()

    def __getstate__(self):
        return {
            "index": self.index,
            "name": self.name,
            "_hash_key": self._hash_key
        }

    def __setstate__(self, state):
        self.index = state['index']
        self.name = state['name']
        self._hash_key = state['_hash_key']


@cython.freelist(1000)
cdef class ComplementFragment(object):
    cdef:
        public double mass
        public list keys

    def __init__(self, mass, keys=None):
        self.mass = mass
        self.keys = keys or []

    def __repr__(self):
        return "{self.__class__.__name__}({self.mass})".format(self=self)


@cython.freelist(10000)
cdef class PartialGlycanSolution(object):
    cdef:
        public double peptide_mass
        public double score
        public set core_matches
        public set fragment_matches
        public int32_t glycan_index

    def __init__(self, peptide_mass=-1, score=0, core_matches=None, fragment_matches=None, int32_t glycan_index=-1):
        if core_matches is None:
            core_matches = set()
        if fragment_matches is None:
            fragment_matches = set()
        self.peptide_mass = peptide_mass
        self.score = score
        self.core_matches = core_matches
        self.fragment_matches = fragment_matches
        self.glycan_index = glycan_index

    @staticmethod
    cdef PartialGlycanSolution _create():
        cdef PartialGlycanSolution self = PartialGlycanSolution.__new__(PartialGlycanSolution)
        self.peptide_mass = -1
        self.score = 0.0
        self.core_matches = set()
        self.fragment_matches = set()
        self.glycan_index = -1
        return self

    def __repr__(self):
        template =("{self.__class__.__name__}({self.peptide_mass}, {self.score}, "
                   "{self.core_matches}, {self.fragment_matches}, {self.glycan_index})")
        return template.format(self=self)

    @property
    def fragment_match_count(self):
        return len(self.fragment_matches)


cdef class GlycanFragmentIndex(object):
    '''A fast and sparse in-memory fragment ion index for quickly matching peaks against multiple
    glycan composition complements.

    Based upon the complement ion indexing strategy described in [1]_.

    Attributes
    ----------
    members : :class:`list` of :class:`GlycanCombinationRecord`
        The glycan composition combinations in this index.
    unique_fragments : :class:`dict` of :class:`str` -> :class:`dict` of :class:`str` -> :class:`IndexGlycanCompositionFragment`
        An internment table for each glycosylation type mapping fragment name to glycan composition fragments
    fragment_index : :class:`dict` of :class:`int` -> :class:`list` of :class:`ComplementFragment`
        A sparse index mapping :attr:`resolution` scaled masses to bins. The mass values and binned fragments are complements of
        true fragments.
    counter : int
        A counter to assign each unique fragment a unique integer index.
    fragment_weight : float
        A scoring parameter to weight overall coverage with.
    core_weight : float
        A scoring parameter to weight core motif coverage with.
    resolution : float
        A scaling factor to convert real masses into truncated bin indices.


    References
    ----------
    ..[1] Zeng, W., Cao, W., Liu, M., He, S., & Yang, P. (2021). Precise, Fast and
      Comprehensive Analysis of Intact Glycopeptides and Monosaccharide-Modifications with pGlyco3.
      Bioarxiv. https://doi.org/https://doi.org/10.1101/2021.02.06.430063
    '''

    cdef:
        public list members
        public object glycosylation_type
        public dict unique_fragments
        public list _fragments
        public dict fragment_index
        public long counter
        public double fragment_weight
        public double core_weight
        public double lower_bound
        public double upper_bound
        public double resolution

    def __init__(self, members=None, fragment_weight=0.56, core_weight=0.42, resolution=100.0):
        self.members = members or []
        self.unique_fragments = dict()
        self._fragments = []
        self.fragment_index = dict()
        self.counter = 0
        self.fragment_weight = fragment_weight
        self.core_weight = core_weight
        self.lower_bound = float('inf')
        self.upper_bound = 0
        self.resolution = resolution

    cpdef IndexGlycanCompositionFragment _intern(self, fragment, glycosylation_type):
        key = str(HashableGlycanComposition(fragment.key))
        try:
            cache_for_type = self.unique_fragments[glycosylation_type]
        except KeyError:
            cache_for_type = self.unique_fragments[glycosylation_type] = dict()
        try:
            return cache_for_type[key]
        except KeyError:
            fragment = IndexGlycanCompositionFragment._create(
                fragment.mass, None, fragment.key, not fragment.is_core)
            cache_for_type[key] = fragment
            fragment.name = key
            fragment.index = self.counter
            self._fragments.append(fragment)
            self.counter += 1
            return fragment

    cpdef long query_width(self, double mass, double error_tolerance):
        return <long>(mass * error_tolerance * self.resolution) + 1

    cpdef long bin_index(self, double mass):
        return <long>(mass * self.resolution)

    def build(self):
        cdef:
            object key
            long n_frags, n_core, j
            double mass, d, low, high
            PyObject* ptemp
            list comp_frags, frags
            size_t frags_i, frags_n, comp_i, comp_n
            IndexGlycanCompositionFragment frag
            ComplementFragment comp
            GlycanCombinationRecordBase member

        self.members.sort(key=lambda x: (x.dehydrated_mass, x.id))
        self.fragment_index.clear()
        j = 0
        low = float('inf')
        high = 0
        for i, member in enumerate(self.members):
            mass = member.dehydrated_mass
            for glycosylation_type in member.glycan_types:
                n_core = 0
                n_frags = 0
                frags = self._get_fragments(member, glycosylation_type)
                for frag_i in range(PyList_GET_SIZE(frags)):
                    frag = self._intern(<object>PyList_GET_ITEM(frags, frag_i), glycosylation_type)
                    if not frag.is_extended:
                        n_core += 1
                    n_frags += 1
                    d = mass - frag.mass
                    if d < 0:
                        d = 0
                    if high < d:
                        high = d
                    if low > d:
                        low = d
                    key = PyInt_FromLong(self.bin_index(d))
                    ptemp = PyDict_GetItem(self.fragment_index, key)
                    if ptemp == NULL:
                        comp_frags = []
                        PyDict_SetItem(self.fragment_index, key, comp_frags)
                    else:
                        comp_frags = <list>ptemp
                    for comp_i in range(PyList_GET_SIZE(comp_frags)):
                        comp = <ComplementFragment>PyList_GET_ITEM(comp_frags, comp_i)
                        if abs(comp.mass - d) <= 1e-3:
                            PyList_Append(comp.keys, (frag, i, glycosylation_type))
                            break
                    else:
                        comp_frags.append(ComplementFragment(d, [(frag, i, glycosylation_type)]))
                        j += 1
                member.fragment_set_properties[glycosylation_type] = (n_core, n_frags)
            member.clear()
        self.lower_bound = low
        self.upper_bound = high
        return j

    cpdef list _get_fragments(self, GlycanCombinationRecordBase gcr, object glycosylation_type):
        if glycosylation_type == GlycanTypes_n_glycan:
            return gcr.get_n_glycan_fragments()
        elif glycosylation_type == GlycanTypes_o_glycan:
            return gcr.get_o_glycan_fragments()
        elif glycosylation_type == GlycanTypes_gag_linker:
            return gcr.get_gag_linker_glycan_fragments()
        else:
            raise ValueError(glycosylation_type)

    @cython.cdivision
    cdef int _match_fragments(self, double delta_mass, DeconvolutedPeak peak, double error_tolerance, dict result):
            cdef:
                object idx, glycosylation_type
                tuple frag_id
                list comp_frags, keys
                dict sols_for_glycan
                long key, width, off
                ComplementFragment comp_frag
                IndexGlycanCompositionFragment frag
                GlycanCombinationRecordBase rec
                PartialGlycanSolution sol
                size_t comp_i, comp_n, key_i, key_n, n_core, n_frag
                PyObject* ptemp

            key = self.bin_index(delta_mass)
            width = <long>(peak.neutral_mass * error_tolerance * self.resolution) + 1

            for off in range(-width, width + 1):
                ptemp = PyDict_GetItem(self.fragment_index, key + off)
                if ptemp == NULL:
                    continue
                else:
                    comp_frags = <list>ptemp
                comp_n = PyList_GET_SIZE(comp_frags)
                for comp_i in range(comp_n):
                    comp_frag = <ComplementFragment>PyList_GET_ITEM(comp_frags, comp_i)
                    if abs(comp_frag.mass - delta_mass) / peak.neutral_mass <= error_tolerance:
                        keys = comp_frag.keys
                        key_n = PyList_GET_SIZE(keys)
                        for key_i in range(key_n):
                            frag_id = <tuple>PyList_GET_ITEM(keys, key_i)
                            frag = <IndexGlycanCompositionFragment>PyTuple_GET_ITEM(frag_id, 0)
                            idx = <object>PyTuple_GET_ITEM(frag_id, 1)
                            glycosylation_type = <object>PyTuple_GET_ITEM(frag_id, 2)

                            # Fetch the dict for the glycan's solutions across glycosylation types.
                            # If it doesn't exist, populate it now.
                            ptemp = PyDict_GetItem(result, idx)
                            if ptemp == NULL:
                                sols_for_glycan = dict()
                                PyDict_SetItem(result, idx, sols_for_glycan)
                            else:
                                sols_for_glycan = <dict>ptemp
                            # From solutions for the current glycan across glycosylation types,
                            # fetch the solution for the matched glycosylation type. If it does not
                            # exist, populate it now.
                            ptemp = PyDict_GetItem(sols_for_glycan, glycosylation_type)
                            if ptemp == NULL:
                                sol = PartialGlycanSolution._create()
                                sol.glycan_index = idx
                                PyDict_SetItem(sols_for_glycan, glycosylation_type, sol)
                            else:
                                sol = <PartialGlycanSolution>ptemp
                            sol.score += math.log10(peak.intensity) * (
                                1 - ((abs(delta_mass - comp_frag.mass) / peak.neutral_mass) / error_tolerance) ** 4)
                            if not frag.is_extended:
                                PySet_Add(sol.core_matches, frag.index)
                            PySet_Add(sol.fragment_matches, frag.index)
            return 0

    @cython.cdivision
    cpdef list match(self, scan, double error_tolerance=1e-5, MassShiftBase mass_shift=Unmodified, query_mass=None):
        cdef:
            double precursor_mass, d
            double best_score
            PartialGlycanSolution sol, best_solution
            dict result, sols_for_glycan
            DeconvolutedPeakSet peak_set
            DeconvolutedPeak peak
            object glycosylation_type
            tuple id_sol, glytype_sol
            list out, solutions
            double coverage, mass_shift_tandem_mass, mass_shift_mass
            GlycanCombinationRecordBase rec
            size_t peak_i, n_peaks, n_core, n_frag, sols_i, sols_n
            PyObject* ptemp

        if query_mass is None:
            precursor_mass = scan.precursor_information.neutral_mass
        else:
            precursor_mass = query_mass

        mass_shift_mass = mass_shift.mass
        mass_shift_tandem_mass = mass_shift.tandem_mass

        peak_set = scan.deconvoluted_peak_set
        n_peaks = peak_set.get_size()
        result = dict()
        for peak_i in range(n_peaks):
            peak = peak_set.getitem(peak_i)
            # Subtract the precursor mass shift mass to match unmodified peaks
            d = (precursor_mass - peak.neutral_mass - mass_shift_mass)
            self._match_fragments(d, peak, error_tolerance, result)
            if mass_shift_tandem_mass != 0:
                # Add back the tandem mass shift to match modified peaks
                d = (precursor_mass - peak.neutral_mass - mass_shift_mass + mass_shift_tandem_mass)
                self._match_fragments(d, peak, error_tolerance, result)

        out = []
        keys = PyDict_Items(result)
        key_n = PyList_GET_SIZE(keys)
        for key_i in range(key_n):
            id_sol = <tuple>PyList_GET_ITEM(keys, key_i)
            idx = <object>PyTuple_GET_ITEM(id_sol, 0)
            sols_for_glycan = <dict>PyTuple_GET_ITEM(id_sol, 1)
            rec = <GlycanCombinationRecordBase>PyList_GET_ITEM(self.members, idx)

            best_score = -1e6
            best_solution = None
            solutions = PyDict_Items(sols_for_glycan)
            sols_n = PyList_GET_SIZE(solutions)
            for sols_i in range(sols_n):
                glytype_sol = <tuple>PyList_GET_ITEM(solutions, sols_i)

                glycosylation_type = <object>PyTuple_GET_ITEM(glytype_sol, 0)
                sol = <PartialGlycanSolution>PyTuple_GET_ITEM(glytype_sol, 1)
                # Reusing container variable lazily
                id_sol = <tuple>PyDict_GetItem(rec.fragment_set_properties, glycosylation_type)
                n_core = PyInt_AsLong(<object>PyTuple_GET_ITEM(id_sol, 0))
                n_frag = PyInt_AsLong(<object>PyTuple_GET_ITEM(id_sol, 1))
                coverage = (PySet_GET_SIZE(sol.core_matches) * 1.0 / n_core) ** self.core_weight * (
                    PySet_GET_SIZE(sol.fragment_matches) * 1.0 / n_frag) ** self.fragment_weight
                sol.score *= coverage
                sol.peptide_mass = precursor_mass - rec.dehydrated_mass - mass_shift_mass
                if best_score < sol.score and sol.peptide_mass > 0:
                    best_score = sol.score
                    best_solution = sol
            if best_solution is not None:
                out.append(best_solution)

        out.sort(key=partial_score_getter, reverse=True)
        return out


cpdef double partial_score_getter(object x):
    return (<PartialGlycanSolution>x).score


import itertools

from glypy.composition import Composition
from glypy.composition.composition_transform import strip_derivatization
from glypy.structure.glycan_composition import FrozenMonosaccharideResidue, MonosaccharideResidue
from glypy.io.nomenclature.identity import is_a

from glycresoft.structure.fragment_match_map import SpectrumGraph
from glycresoft.structure.denovo import MassWrapper, PathFinder


class DeNovoSolution(object):
    def __init__(self, scan, supporting_paths=None, completion_gap=None):
        self.scan = scan
        self.supporting_paths = supporting_paths
        self.completion_gap = completion_gap

    @property
    def precursor_mass(self):
        return self.scan.precursor_information.neutral_mass


class DeNovoGlycanSequencer(PathFinder):
    def __init__(self, spectrum, components, precursor_error_tolerance=1e-5, product_error_tolerance=2e-5):
        super(DeNovoGlycanSequencer, self).__init__(spectrum, components, product_error_tolerance)
        self.precursor_error_tolerance = precursor_error_tolerance



import numpy as np

from glycresoft.structure import FragmentMatchMap

from .base import (
    PeptideSpectrumMatcherBase, ChemicalShift,
    HCDFragmentationStrategy, IonSeries)


class SimpleCountScorer(PeptideSpectrumMatcherBase):
    def __init__(self, scan, sequence, mass_shift=None):
        super(SimpleCountScorer, self).__init__(scan, sequence, mass_shift)
        self._score = None
        self.solution_map = FragmentMatchMap()

    def calculate_score(self, **kwargs):
        self._score = len(self.solution_map)
        return self._score


class SimpleCoverageScorer(PeptideSpectrumMatcherBase):
    def __init__(self, scan, sequence, mass_shift=None):
        super(SimpleCoverageScorer, self).__init__(scan, sequence, mass_shift)
        self._score = None
        self.solution_map = FragmentMatchMap()

    def _compute_coverage_vectors(self):
        n_term_ions = np.zeros(len(self.target))
        c_term_ions = np.zeros(len(self.target))

        for frag in self.solution_map.fragments():
            if frag.series in (IonSeries.b, IonSeries.c):
                n_term_ions[frag.position] = 1
            elif frag.series in (IonSeries.y, IonSeries.z):
                c_term_ions[frag.position] = 1
        n_term_ions[0] = 0
        return n_term_ions, c_term_ions

    def compute_coverage(self):
        (n_term_ions, c_term_ions) = self._compute_coverage_vectors()

        mean_coverage = np.mean(np.log2(n_term_ions + c_term_ions[::-1] + 1) / np.log2(3))

        return mean_coverage

    def calculate_score(self, **kwargs):
        score = self._coverage_score()
        self._score = score
        return score

    def _coverage_score(self):
        return self.compute_coverage()



try:
    _has_c = True
    from glycresoft._c.tandem.tandem_scoring_helpers import _peptide_compute_coverage_vectors, compute_coverage
    SimpleCoverageScorer._compute_coverage_vectors = _peptide_compute_coverage_vectors
    SimpleCoverageScorer.compute_coverage = compute_coverage
except ImportError:
    _has_c = False



from collections import defaultdict

from . import (smooth_overlaps, build_rt_interval_tree)


def binary_search_with_flag(array, mass, error_tolerance=1e-5):
    """Binary search an ordered array of objects with :attr:`neutral_mass`
    using a PPM error tolerance of `error_toler

    Parameters
    ----------
    array : list
        An list of objects, sorted over :attr:`neutral_mass` in increasing order
    mass : float
        The mass to search for
    error_tolerance : float, optional
        The PPM error tolerance to use when deciding whether a match has been found

    Returns
    -------
    int:
        The index in `array` of the best match
    bool:
        Whether or not a match was actually found, used to
        signal behavior to the caller.
    """
    lo = 0
    n = hi = len(array)
    while hi != lo:
        mid = (hi + lo) // 2
        x = array[mid]
        err = (x.neutral_mass - mass) / mass
        if abs(err) <= error_tolerance:
            best_index = mid
            best_error = abs(err)
            i = mid - 1
            while i >= 0:
                x = array[i]
                err = abs((x.neutral_mass - mass) / mass)
                if err < best_error:
                    best_error = err
                    best_index = i
                elif err > error_tolerance:
                    break
                i -= 1

            i = mid + 1
            while i < n:
                x = array[i]
                err = abs((x.neutral_mass - mass) / mass)
                if err < best_error:
                    best_error = err
                    best_index = i
                elif err > error_tolerance:
                    break
                i += 1
            return best_index, True
        elif (hi - lo) == 1:
            return mid, False
        elif err > 0:
            hi = mid
        elif err < 0:
            lo = mid
    return 0, False


class ChromatogramIndex(object):
    """An ordered collection of Chromatogram-like objects with fast searching
    and features. Supports Sequence operations.

    Attributes
    ----------
    chromatograms: list of Chromatogram
        list of chromatogram-like objects, ordered by neutral mass
    """
    def __init__(self, chromatograms, sort=True):
        if sort:
            self._sort(chromatograms)
        else:
            self.chromatograms = list(chromatograms)

    def _sort(self, iterable):
        self.chromatograms = sorted([c for c in iterable if len(c)], key=lambda x: (x.neutral_mass, x.start_time))

    def add(self, chromatogram, sort=True):
        self.chromatograms.append(chromatogram)
        if sort:
            self._sort(self.chromatograms)

    def __iter__(self):
        return iter(self.chromatograms)

    def __getitem__(self, i):
        return self.chromatograms[i]

    def __len__(self):
        return len(self.chromatograms)

    def find_mass(self, mass, ppm_error_tolerance=1e-5):
        index, flag = self._binary_search(mass, ppm_error_tolerance)
        if flag:
            return self[index]
        else:
            return None

    def find_all_by_mass(self, mass, ppm_error_tolerance=1e-5):
        if len(self) == 0:
            return self.__class__([], sort=False)
        center_index, _ = self._binary_search(mass, ppm_error_tolerance)
        low_index = center_index
        while low_index > 0:
            x = self[low_index - 1]
            if abs((mass - x.neutral_mass) / x.neutral_mass) > ppm_error_tolerance:
                break
            low_index -= 1

        n = len(self)
        high_index = center_index - 1
        while high_index < n - 1:
            x = self[high_index + 1]
            if abs((mass - x.neutral_mass) / x.neutral_mass) > ppm_error_tolerance:
                break
            high_index += 1
        if low_index == high_index == center_index:
            x = self[center_index]
            if abs((mass - x.neutral_mass) / x.neutral_mass) > ppm_error_tolerance:
                return self.__class__([], sort=False)
        items = self[low_index:high_index + 1]
        items = [
            c for c in items if abs(
                (c.neutral_mass - mass) / mass) < ppm_error_tolerance
        ]
        return self.__class__(items, sort=False)

    def mass_between(self, low, high):
        n = len(self)
        if n == 0:
            return self.__class__([])
        low_index, flag = self._binary_search(low, 1e-5)
        low_index = max(0, min(low_index, n - 1))
        if self[low_index].neutral_mass < low:
            low_index += 1
        high_index, flag = self._binary_search(high, 1e-5)
        high_index += 2
        # high_index = min(n - 1, high_index)
        if (high_index < n) and (self[high_index].neutral_mass > high):
            high_index -= 1
        items = self[low_index:high_index]
        items = [c for c in items if low <= c.neutral_mass <= high]
        return self.__class__(items, sort=False)

    def _binary_search(self, mass, error_tolerance=1e-5):
        return binary_search_with_flag(self, mass, error_tolerance)

    def __repr__(self):
        return repr(list(self))

    def _repr_pretty_(self, p, cycle):
        return p.pretty(list(self))

    def __str__(self):
        return str(list(self))


try:
    _ChromatogramIndex = ChromatogramIndex
    from glycresoft._c.chromatogram_tree.index import ChromatogramIndex
except ImportError:
    pass


class ChromatogramFilter(ChromatogramIndex):
    """An ordered collection of Chromatogram-like objects with fast searching
    and filtering features. Supports Sequence operations.

    Attributes
    ----------
    chromatograms: list of Chromatogram
        list of chromatogram-like objects, ordered by neutral mass
    key_map: dict
        A mapping between values appearing in the :attr:`Chromatogram.key` and :class:`DisjointChromatogramSet`
        instances containing all occurrences of that key, ordered by time.
    rt_interval_tree: IntervalTreeNode
        An interval tree over retention time containing all of the chromatograms in
        :attr:`chromatograms`
    """
    def __init__(self, chromatograms, sort=True):
        super(ChromatogramFilter, self).__init__(chromatograms, sort=sort)
        self._key_map = None
        self._intervals = None

    def _invalidate(self):
        self._key_map = None
        self._intervals = None

    def _build_key_map(self):
        self._key_map = defaultdict(list)
        for chrom in self:
            self._key_map[chrom.key].append(chrom)
        for key in tuple(self._key_map.keys()):
            self._key_map[key] = DisjointChromatogramSet(self._key_map[key])
        return self._key_map

    def _build_rt_interval_tree(self):
        self._intervals = build_rt_interval_tree(self)

    @property
    def key_map(self):
        if self._key_map is None:
            self._build_key_map()
        return self._key_map

    @property
    def rt_interval_tree(self):
        if self._intervals is None:
            self._build_rt_interval_tree()
        return self._intervals

    def find_all_instances(self, key):
        return self.key_map[key]

    def find_key(self, key):
        try:
            return self.key_map[key][0]
        except (KeyError, IndexError):
            return None

    def min_points(self, n=3, keep_if_msms=True):
        self.chromatograms = [c for c in self if (len(c) >= n) or c.has_msms]
        return self

    def split_sparse(self, delta_rt=1.):
        self.chromatograms = [
            seg for c in self
            for seg in c.split_sparse(delta_rt)
        ]
        return self

    def spanning(self, rt):
        return self.__class__((c for c in self if c.start_time <= rt <= c.end_time), sort=False)

    def contained_in_interval(self, start, end):
        return self.__class__(
            (c for c in self if ((c.start_time <= start and c.end_time >= start) or (
                c.start_time >= start and c.end_time <= end) or (
                c.start_time >= start and c.end_time >= end and c.start_time <= end) or (
                c.start_time <= start and c.end_time >= start) or (
                c.start_time <= end and c.end_time >= end))), sort=False)

    def after(self, t):
        out = []
        for c in self:
            c = c.clone()
            c.truncate_before(t)
            if len(c) > 0:
                out.append(c)
        return self.__class__(out, sort=False)

    def before(self, t):
        out = []
        for c in self:
            c = c.clone()
            c.truncate_after(t)
            if len(c) > 0:
                out.append(c)
        return self.__class__(out, sort=False)

    def filter(self, filter_fn):
        return self.__class__([x for x in self if filter_fn(x)], sort=False)

    @classmethod
    def process(cls, chromatograms, min_points=5, percentile=10, delta_rt=1.):
        return cls(chromatograms).split_sparse(delta_rt).min_points(min_points)

    def smooth_overlaps(self, mass_error_tolerance=1e-5):
        return self.__class__(smooth_overlaps(self, mass_error_tolerance))

    def extend(self, other):
        chroma = []
        chroma.extend(self)
        chroma.extend(other)
        self.chromatograms = [c for c in sorted([c for c in chroma if len(c)], key=lambda x: (
                              x.neutral_mass, x.start_time))]
        self._invalidate()

    def __add__(self, other):
        inst = self.__class__([])
        inst.extend(self)
        inst.extend(other)
        return inst


class DisjointChromatogramSet(object):
    def __init__(self, chromatograms):
        self.group = sorted(chromatograms, key=lambda c: c.start_time)

    def linear_search(self, start_time, end_time):
        center_time = (start_time + end_time) / 2.
        for chrom in self.group:
            if chrom.start_time <= center_time <= chrom.end_time:
                return chrom

    def find_overlap(self, chromatogram):
        return self.linear_search(
            chromatogram.start_time,
            chromatogram.end_time)

    def replace(self, original, replacement):
        i = self.group.index(original)
        self.group[i] = replacement

    def __getitem__(self, i):
        return self.group[i]

    def __iter__(self):
        return iter(self.group)

    def __repr__(self):
        return repr(list(self))

    def _repr_pretty_(self, p, cycle):
        return p.pretty(self.group)

    def __str__(self):
        return str(list(self))

    def __len__(self):
        return len(self.group)



import unittest

import glypy
from glycopeptidepy import PeptideSequence
from ms_deisotope.output import ProcessedMSFileLoader

from glycresoft.tandem.glycopeptide import core_search
from glycresoft.tandem.glycopeptide.core_search import (
    GlycanCombinationRecord, GlycanTypes, GlycanFilteringPeptideMassEstimator)

from .fixtures import get_test_data


peptide_mass = PeptideSequence("YLGNATAIFFLPDEGK").mass
gc1 = glypy.glycan_composition.HashableGlycanComposition.parse("{Hex:5; HexNAc:4; Neu5Ac:1}")
gc2 = glypy.glycan_composition.HashableGlycanComposition.parse("{Hex:5; HexNAc:4; Neu5Ac:2}")
gc3 = glypy.glycan_composition.HashableGlycanComposition.parse("{Hex:6; HexNAc:5; Neu5Ac:2}")
glycan_compositions = [gc1, gc2, gc3]

glycan_database = []
for i, gc in enumerate(glycan_compositions):
    record = GlycanCombinationRecord(
        i + 1, gc.mass() - gc.composition_offset.mass, gc, 1, [
            GlycanTypes.n_glycan,
            GlycanTypes.o_glycan,

        ])
    glycan_database.append(record)



class TestGlycanFilteringPeptideMassEstimator(unittest.TestCase):
    def load_spectra(self):
        return list(ProcessedMSFileLoader(get_test_data("example_glycopeptide_spectra.mzML")))

    def make_estimator(self):
        return GlycanFilteringPeptideMassEstimator(glycan_database)

    def test_estimate(self):
        estimator = self.make_estimator()
        scans = self.load_spectra()
        scan = scans[0]
        ranked = estimator.match(scan)
        match = ranked[0]
        print(match)
        self.assertAlmostEqual(match.score, 29.715553766294754, 3)


if __name__ == "__main__":
    unittest.main()




cdef class MassObject(object):
    cdef:
        public object obj
        public double mass


cdef class NeutralMassDatabaseImpl(object):
    cdef:
        public list structures
        public object mass_getter

    cdef list _prepare(self, list structures, bint sort=*)
    cpdef size_t search_binary(self, double mass, double error_tolerance=*)

    cdef size_t get_size(self)
    cdef object get_item(self, size_t i)
    cpdef list search_mass(self, double mass, double error_tolerance=*)
    cpdef object search_between(self, double lower, double higher)